#include <stdlib.h>
#include <string.h>

#include <inc/mem.h>
#include <inc/dos.h>
#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/sel.h>
#include <inc/red.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "INT"
static int log_level;

#define PCREL(at, to) do{*(ui*)(at)=(ui)(to)-(ui)(at)-4;}while(0)

static int_handler int_tab[256];
static int enabled;

void INT_log_regs (INT_ARGS)
{
  log ("intregs:\n"
       "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
       "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
       "eip=%#010x efl=%#010x\n"
       "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
       "fs=%#06x gs=%#06x\n",
       EAX(r), EBX(r), ECX(r), EDX(r), ESI(r), EDI(r), EBP(r),
       ESP(r), EIP(r), EFL(r),
       CS (r), DS(r), ES(r), FS (r), ASM_ss, GS (r)
    );
}

static void unhandled_int (INT_ARGS)
{
  INT_log_regs (r);
  Fail ("exiting");
}

void INT_enable ()
{
  linfo ("%p\n", __builtin_return_address (0));
  enabled = 1;
}

void INT_disable ()
{
  linfo ("%p\n", __builtin_return_address (0));
  enabled = 0;
}

int INT_is_enabled ()
{
  return enabled;
}

static void get_rm (int nint, us *seg, us *off)
{
  int e;
  us *p;

  p = (us *) MEM_REAL (nint * 4);
  e = enabled;
  enabled = 0;
  *off = p[0];
  *seg = p[1];
  enabled = e;
}

static void set_rm (int nint, us seg, us off)
{
  int e;
  us *p;

  p = (us *) MEM_REAL (nint * 4);
  e = enabled;
  enabled = 0;
  p[0] = off;
  p[1] = seg;
  enabled = e;
}

void INT_get_rm (int nint, us *sel, us *off)
{
  get_rm (nint, sel, off);
}

void INT_set_rm (int nint, us sel, us off)
{
  set_rm (nint, sel, off);
}

void INT_set (int nint, int_handler handler)
{
  int e;

  e = enabled;
  enabled = 0;
  int_tab[nint] = handler;
  enabled = e;
}

int_handler INT_get (int nint)
{
  return int_tab[nint];
}

static void ignore (INT_ARGS)
{
  lwarn ("ignoring\n");
  INT_log_regs (r);
}

void INT_init ()
{
  int i;
  ub *redir;

  DEREF (ignore);
  DEREF (unhandled_int);

  log_level = MUP_get_log_level ("int");

  redir = (ub *) RED_FLAT (RED_INTERRUPT_00);

  for (i = 0; i < 256; i++) {
    INT_set_rm (i, RED_SEGMENT, RED_INTERRUPT_00 + i);
    redir[i] = RED_ICE;
  }

  INT_set_rm (0x66, 0, 0);
  INT_set_rm (0x67, 0, 0);
  INT_set_rm (0x68, 0, 0);
  INT_set_rm (0x69, 0, 0);

  redir[0x0d] = 0xcf;
  redir[0x0f] = 0xcf;
  redir[0x30] = 0xcf;
  redir[0x1c] = 0xcf;
  redir[0x67] = 0xcf;

  INT_set (0x10, I10);
  INT_set (0x11, I11);
  INT_set (0x12, I12);
  INT_set (0x15, I15);
  INT_set (0x16, I16);
  INT_set (0x1A, I1A);
  INT_set (0x21, I21);
  INT_set (0x29, I29);
  INT_set (0x2f, I2F);
  INT_set (0x33, I33);

  enabled = 1;
}
