#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>

#include <inc/tid.h>
#include <inc/eio.h>
#include <inc/pie.h>
#include <inc/dos.h>
#include <inc/ldt.h>
#include <inc/asm.h>
#include <inc/mem.h>
#include <inc/sel.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/pit.h>
#include <inc/pic.h>
#include <inc/rtc.h>
#include <inc/vid.h>
#include <inc/evt.h>
#include <inc/x86.h>
#include <inc/red.h>
#include <inc/sou.h>
#include <inc/xms.h>
#include <inc/log.h>
#include <inc/mup.h>

#include <sys/mman.h>

#include <qemu.h>

/* extern int cpu_signal_handler (int host_signum, struct siginfo *info, void *puc); */
extern int page_unprotect (unsigned long address);

#undef R_EAX
#undef R_EBX
#undef R_ECX
#undef R_EDX
#undef R_ESI
#undef R_EDI
#undef R_EBP
#undef R_ESP
#undef R_CS
#undef R_DS
#undef R_ES
#undef R_SS
#undef R_FS
#undef R_GS

enum {
  R_EAX = 0,
  R_ECX = 1,
  R_EDX = 2,
  R_EBX = 3,
  R_ESP = 4,
  R_EBP = 5,
  R_ESI = 6,
  R_EDI = 7,

  R_ES = 0,
  R_CS = 1,
  R_SS = 2,
  R_DS = 3,
  R_FS = 4,
  R_GS = 5,
};

static int log_level;
#define LOG_PREFIX "SIG"

#define SIGAX(r) GETLOWORD(SIGEAX(r))
#define SIGBX(r) GETLOWORD(SIGEBX(r))
#define SIGCX(r) GETLOWORD(SIGECX(r))
#define SIGDX(r) GETLOWORD(SIGEDX(r))
#define SIGSI(r) GETLOWORD(SIGESI(r))
#define SIGDI(r) GETLOWORD(SIGEDI(r))
#define SIGBP(r) GETLOWORD(SIGEBP(r))

static int r_lock;
#define r_lock() do{if(r_lock)abort();else r_lock=1;}while(0)
#define r_unlock() do{if(!r_lock)abort();else r_lock=0;}while(0)

struct unicontext {
  int signum;
  int trapno;
  int err;
  int cr2;
  sigset_t set;
  regs86 *regs;
};

union sigunion { regs86 regs; ub data[CONTEXT_SZ]; };
struct cpu_context {
  CPUState *env;
  int async;
};

static struct {
  int x86, sig, pic_loop;
} in;

CPUState *global_env;
FILE *logfile;
int loglevel;
CPUState *cpu_single_env;

static int need_confirmation;
static int outside;
static int loop_count;
static int vif;
static int debugger_present = -1;
static char *debugger_command;
static union sigunion sigctx;
static struct SIG_mouse_info minfo;

#define SETCFL(env, cfl) \
env->eflags = (env->eflags & 0xffff0000) | ((cfl) & 0xffff)
#define QEAX(r) (r)->regs[R_EAX]
#define QEBX(r) (r)->regs[R_EBX]
#define QECX(r) (r)->regs[R_ECX]
#define QEDX(r) (r)->regs[R_EDX]
#define QESI(r) ((r)->regs[R_ESI])
#define QEDI(r) (r)->regs[R_EDI]
#define QEBP(r) (r)->regs[R_EBP]
#define QESP(r) (r)->regs[R_ESP]

#define QCS(r) (r)->segs[R_CS].selector
#define QDS(r) ((r)->segs[R_DS].selector)
#define QES(r) (r)->segs[R_ES].selector
#define QSS(r) (r)->segs[R_SS].selector
#define QFS(r) (r)->segs[R_FS].selector
#define QGS(r) (r)->segs[R_GS].selector

#define QEIP(r) (r)->eip
#define QEFL(r) (r)->eflags

#define QIP(r) (QEIP(r)&0xffff)
#define QAH(r) ((QEAX(r)>>8)&0xff)
#define QAX(r) (QEAX(r)&0xffff)
#define QDX(r) (QEDX(r)&0xffff)
#define QSI(r) (QESI(r)&0xffff)
#define QDI(r) (QEDI(r)&0xffff)
#define QCX(r) (QECX(r)&0xffff)
#define QAL(r) (QEAX(r)&0xff)
#define QBL(r) (QEBX(r)&0xff)
#define SET_QAX(r, v) QEAX(r)=(QEAX(r)&0xffff0000)|((v)&0xffff)
#define SET_QBX(r, v) QEBX(r)=(QEBX(r)&0xffff0000)|((v)&0xffff)
#define SET_QDX(r, v) QEDX(r)=(QEDX(r)&0xffff0000)|((v)&0xffff)
#define SET_QCX(r, v) QECX(r)=(QECX(r)&0xffff0000)|((v)&0xffff)
#define SET_QSI(r, v) QESI(r)=(QESI(r)&0xffff0000)|((v)&0xffff)
#define SET_QDI(r, v) QEDI(r)=(QEDI(r)&0xffff0000)|((v)&0xffff)
#define SET_QAL(r, v) QEAX(r)=(QEAX(r)&0xffffff00)|((v)&0xff)
#define SET_QCL(r, v) QECX(r)=(QECX(r)&0xffffff00)|((v)&0xff)

static void copy_context (void *dst, void *src);

static int is_debugger_present ()
{
  FILE *f;
  char buf[256];
  pid_t ppid;

  ppid = getppid ();
  snprintf (buf, sizeof (buf), "/proc/%d/cmdline", ppid);

  f = fopen (buf, "r");
  if (NULL == f)
    return 0;

  fgets (buf, 256, f);
  fclose (f);

  return NULL != strstr (buf, "gdb");
}

static int attach_debugger ()
{
  int err;
  pid_t pid, cpid;
  char buf[8];
  char *argv[3];

  if (debugger_command) {
    pid = getpid ();
    snprintf (buf, sizeof (buf), "%#x", pid);

    cpid = fork ();
    if (-1 == cpid) {
      err = errno;
      log ("fork failed\n"
           "syserr: `%s'\n", strerror (err));
      return -1;
    }

    argv[0] = debugger_command;
    argv[1] = buf;
    argv[2] = NULL;

    if (0 == cpid) {
      if (-1 == execve (debugger_command, argv, environ)) {
        err = errno;
        log ("execl failed\nsyserr: `%s'\n", strerror (err));
      }
      exit (EXIT_FAILURE);
    }

    return cpid;
  }
  else {
    return -1;
  }
}

static void log_qemu_regs (CPUX86State *env)
{
  char m[X86_EFLAGS_BUF_SIZE];

  log ("QEMU regs:\n"
       "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
       "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
       "eip=%#010x efl=%#010x %s\n"
       "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
       "fs=%#06x gs=%#06x\n",
       env->regs[R_EAX],
       env->regs[R_EBX],
       env->regs[R_ECX],
       env->regs[R_EDX],
       env->regs[R_ESI],
       env->regs[R_EDI],
       env->regs[R_EBP],
       env->regs[R_ESP],
       env->eip,
       env->eflags, X86_efltoa (env->eflags, m),
       env->segs[R_CS].selector,
       env->segs[R_DS].selector,
       env->segs[R_ES].selector,
       env->segs[R_SS].selector,
       env->segs[R_FS].selector,
       env->segs[R_GS].selector);
}


void SIG_log_regs (regs86 *r)
{
  char m[X86_EFLAGS_BUF_SIZE];

  log ("sigregs:\n"
       "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
       "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
       "eip=%#010x efl=%#010x %s\n"
       "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
       "fs=%#06x gs=%#06x\n",
       (int) SIGEAX(r), (int) SIGEBX(r), (int) SIGECX(r), (int) SIGEDX(r),
       (int) SIGESI(r), (int) SIGEDI(r), (int) SIGEBP(r), (int) SIGESP(r),
       (int) SIGEIP(r), (int) SIGEFL(r), X86_efltoa (SIGEFL (r), m),
       SIGCS(r), SIGDS(r), SIGES(r), SIGSS(r), SIGFS(r), SIGGS(r)
    );

  log_qemu_regs (global_env);

  if (in.x86) {
    X86_dump_current ();
  }
}

static void keeping ()
{
  SOU_run ();

  if (0 == in.pic_loop) {
    PIT_run ();
    RTC_run ();
  }

  if (0 == r_lock) {
    r_lock = 1;
    VID_run ();
    EVT_run ();
    r_lock = 0;
  }
}

void SIG_break (int fatal)
{
  int was_blocked;

  was_blocked = SIG_block ();
  VID_windowed ();

  if (-1 == debugger_present)
    exit (EXIT_FAILURE);

  if (debugger_present) {
    BPT;
  }
  else {
    int status;
    pid_t pid;

    pid = attach_debugger ();
    if (-1 == pid) {
      exit (EXIT_FAILURE);
    }
    if (fatal) waitpid (pid, &status, 0);
  }

  if (fatal) {
    exit (EXIT_FAILURE);
  }

  SIG_unblock (was_blocked);
}

void SIG_wait ()
{
  sigset_t set;

  sigprocmask (SIG_BLOCK, NULL, &set);
  Fail_if (sigismember (&set, SIGALRM));
  sigsuspend (&set);
}

int SIG_block ()
{
  sigset_t set;
  sigset_t old_set;

  sigemptyset (&set);
  sigaddset (&set, SIGALRM);
  sigprocmask (SIG_BLOCK, &set, &old_set);
  return sigismember (&old_set, SIGALRM);
}

void SIG_unblock (int was_blocked)
{
  sigset_t set;

  if (0 == was_blocked) {
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigprocmask (SIG_UNBLOCK, &set, 0);
  }
}

int SIG_is_in_pic_loop ()
{
  return in.pic_loop;
}

void SIG_mouse_callback (struct SIG_mouse_info *info)
{
  memcpy (&minfo, info, sizeof (minfo));
  minfo.pending = 1;
}

static void jump (regs86 *r)
{
  copy_context (r, sigctx.data);
  lverbose ("jumping to pc %#06x %#010lx sp %#06x %#010lx\n",
            SIGCS (&sigctx.regs), SIGEIP (&sigctx.regs),
            SIGSS (&sigctx.regs), SIGESP (&sigctx.regs));
}

static void Cpu_loop (struct cpu_context *ctx);
static void cpu_redir ()
{
  struct cpu_context ctx;

  ctx.env = global_env;
  ctx.async = 1;
  Cpu_loop (&ctx);
  raise (SIGUSR1);
}

static void maybe_service_irqs (CPUState *env, regs86 *r)
{
  int pending, enabled;

  enabled = env->eflags & IF_MASK;
  pending = PIC_irq_pending ();

  ldebug ("in_pic_loop %d, enabled %d vif %d, outside %d, pending %d\n",
          in.pic_loop, enabled, vif, outside, pending);

  if (vif && enabled && pending && (0 == in.pic_loop)) {
    if (outside) {
      in.pic_loop = 1;
      copy_context (sigctx.data, r);
      SIGEIP (r) = (ui) &cpu_redir;
    }
    else {
      cpu_x86_interrupt (env, CPU_INTERRUPT_HARD);
    }
  }
}

static int is_fault_freq_high ()
{
  static int nfaults;
  static ul last_fault;
  ul delta;

  delta = TID_get_abs_nsec () - last_fault;
  nfaults++;

  if (delta > ((ul) TID_NSEC) / 100) {
    int high;

    if (debugger_present)
      high = nfaults > 400;
    else
      high = nfaults > 1000;
    nfaults = 0;
    last_fault += delta;
    return high;
  }

  return 0;
}

void SIG_vm86 (struct vm86_regs *regs, int no_stack)
{
  us *_sp;
  us ss, sp;
  struct cpu_context ctx;

  Fail_if (0 == no_stack);

  ss = RED_SEGMENT + ((RED_END + 0xf) >> 4);
  sp = 0x200 - 6;
  _sp = (us *) ((ss << 4) + sp);
  _sp[0] = RED_SEGMENT;
  _sp[1] = RED_ASYNC_IRET;
  _sp[2] = 0;

  global_env->regs[R_EAX] = regs->eax;
  global_env->regs[R_EBX] = regs->ebx;
  global_env->regs[R_ECX] = regs->ecx;
  global_env->regs[R_EDX] = regs->edx;
  global_env->regs[R_ESI] = regs->esi;
  global_env->regs[R_EDI] = regs->edi;
  global_env->regs[R_EBP] = regs->ebp;
  global_env->regs[R_ESP] = sp;

  global_env->eip = regs->eip;
  global_env->eflags = regs->eflags;

  cpu_x86_load_seg (global_env, R_CS, regs->cs);
  cpu_x86_load_seg (global_env, R_DS, regs->ds);
  cpu_x86_load_seg (global_env, R_ES, regs->es);
  cpu_x86_load_seg (global_env, R_FS, regs->fs);
  cpu_x86_load_seg (global_env, R_GS, regs->gs);
  cpu_x86_load_seg (global_env, R_SS, ss);

  global_env->idt.base = (uint8_t *) MEM_REAL (0);
  global_env->idt.limit = 256 * 4 - 1;

  ctx.env = global_env;
  ctx.async = 0;
  vif = 0;
  Cpu_loop (&ctx);
}

static void set_stack (int size)
{
  void *ptr;
  int res;
  struct sigaltstack ss;

  if (size <= 0)
    return;

  ptr = malloc (size);
  if (NULL == ptr) {
    Fail ("failed to allocate %d bytes for signal stack", size);
  }
  ss.ss_sp = ptr;
  ss.ss_size = size;
  ss.ss_flags = 0;

  res = sigaltstack (&ss, 0);
  if (-1 == res) {
    ERRFail ("can not set signal stack");
  }
}

static void runinterp (regs86 *r, sigset_t *oldset)
{
  int confirm;
  sigset_t set;
  int threshold;
  x86exit_e exit;
  int alarm_blocked;

  in.x86 = 1;
  confirm = need_confirmation;
  need_confirmation = 0;

  threshold = 15000;
  alarm_blocked = sigismember (oldset, SIGALRM);
  if (0 == alarm_blocked) {
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigprocmask (SIG_UNBLOCK, &set, NULL);
  }

  if (confirm) {
    exit = X86_confirm32 (r, threshold);
  }
  else {
    exit = X86_run32 (r, threshold);
  }

  if (0 == alarm_blocked) {
    sigprocmask (SIG_BLOCK, &set, NULL);
  }

  in.x86 = 0;

  switch (exit) {
    case X86_EXIT_TEMPORARY:
      need_confirmation = 1;
      break;

    case X86_EXIT_REQUESTED:
      if ((0 == alarm_blocked) && (0 == need_confirmation)) {
        maybe_service_irqs (global_env, r);
      }
      break;

    case X86_EXIT_BAD_INSN:
    case X86_EXIT_THRESHOLD:
      break;
  }
}

static void signal_failed (struct unicontext *ctx)
{
  log ("Failed to handle signal %d `%s'\n"
       "trap: %#x, error: %#x, cr2: %#x\n",
       ctx->signum, strsignal (ctx->signum),
       ctx->trapno, ctx->err, ctx->cr2);
  SIG_log_regs (ctx->regs);
  SIG_break (1);
}

static void unihandler (struct unicontext *ctx)
{
  regs86 *r;

  r = ctx->regs;
  switch (ctx->signum) {
    case SIGSEGV:
    case SIGTRAP:
    {
      switch (ctx->trapno) {
        case 0x01:
          if (need_confirmation) {
            runinterp (ctx->regs, &ctx->set);
          }
          else signal_failed (ctx);
          break;

#if 1
        case 0x0e:
        {
          ui tmp;

          tmp = ctx->cr2 - MEM_REAL (0xa0000);
          if (tmp < 0x20000) {
            runinterp (ctx->regs, &ctx->set);
          }
          else {
            log ("Page fault\n");
            goto cpu;
          }
          break;
        }
#endif

        default:
          goto cpu;
      }
      break;
    }

    case SIGUSR1:
      in.pic_loop = 0;
      INT_enable ();
      jump (r);
      break;

    case SIGALRM:
      keeping ();
      if (in.x86) {
        X86_stop ();
      }
      else {
        maybe_service_irqs (global_env, r);
      }
      break;

    default:
  cpu:
    {
      int ret;
      /* HERE BE DRAGONS */
      struct siginfo info;
      struct ucontext ucontext;

      info.si_addr = (void *) ctx->cr2;
      memcpy (&ucontext.uc_sigmask, &ctx->set, sizeof (ctx->set));
      ucontext.uc_mcontext.gregs[REG_TRAPNO] = ctx->trapno;
      ucontext.uc_mcontext.gregs[REG_ERR] = ctx->err;
      ucontext.uc_mcontext.gregs[REG_EIP] = SIGEIP (r);

      __asm__ (
        "push %%ebp\n\t"
        "push %0\n\t"
        "push %1\n\t"
        "push %2\n\t"
        "call cpu_x86_signal_handler\n\t"
        "add $12, %%esp\n\t"
        "pop %%ebp\n\t"
        : "=a" (ret)
        : "g" (&ucontext)
        , "g" (&info)
        , "g" (ctx->signum));

      /* ret = cpu_signal_handler (ctx->signum, &info, &ucontext); */

      log ("cr2=%x ret=%d\n", ctx->cr2, ret);
      if (ctx->cr2 < 0)
        abort ();

      if (0 == ret)
        signal_failed (ctx);
      break;
    }
  }
}

#ifdef SIGNALS_PEDAL_TO_METAL

#include <unistd.h>
#include <sys/syscall.h>

typedef void (*sigaction_proto) (int, struct sigcontext_struct);

typedef unsigned long old_sigset_t;
struct y_sigaction {
  __sighandler_t my_sa_handler;
  old_sigset_t sa_mask;
  unsigned long sa_flags;
  void (*sa_restorer)(void);
};

static void copy_context (void *dst, void *src)
{
  struct sigcontext_struct *src_ctx;
  struct _fpstate *fpstate;

  src_ctx = (struct sigcontext_struct *) src;
  memcpy (dst, src, REGS86_SZ);

  fpstate = (struct _fpstate *) ((ui) dst + REGS86_SZ);
  Fail_if (NULL == fpstate);
  ((struct sigcontext_struct *) dst)->fpstate = fpstate;
  memcpy (fpstate, src_ctx->fpstate, sizeof (*fpstate));

  leverbose ("copying context pc %#06x %#010lx sp %#06x %#010lx\n",
             SIGCS (((struct sigcontext_struct *) src)),
             SIGEIP (((struct sigcontext_struct *) src)),
             SIGSS (((struct sigcontext_struct *) src)),
             SIGESP (((struct sigcontext_struct *) src)));
}

static void sig_fatal (int signum, struct sigcontext_struct context)
{
  struct unicontext ctx;
  ASM_restore_fs_gs ();

  sigemptyset (&ctx.set);
  ctx.set.__val[0] = context.oldmask;
  ctx.signum = signum;
  ctx.err = context.err;
  ctx.trapno = context.trapno;
  ctx.cr2 = context.cr2;
  ctx.regs = &context;

  signal (SIGABRT, SIG_DFL);
  signal_failed (&ctx);
}

static void sig_handler (int signum, struct sigcontext_struct context)
{
  struct unicontext ctx;
  ASM_restore_fs_gs ();

  sigemptyset (&ctx.set);
  ctx.set.__val[0] = context.oldmask;
  ctx.signum = signum;
  ctx.err = context.err;
  ctx.trapno = context.trapno;
  ctx.cr2 = context.cr2;
  ctx.regs = &context;

  unihandler (&ctx);
}

static int set_handler (int signo, sigaction_proto handler)
{
  struct y_sigaction sa;

  sigset_t set;

  sigemptyset (&set);
  sigaddset (&set, SIGALRM);

  sa.sa_mask = set.__val[0];
  sa.my_sa_handler = (__sighandler_t) handler;
  sa.sa_flags = SA_RESTART | SA_ONSTACK;
  sa.sa_restorer = 0;
  syscall (SYS_sigaction, signo, &sa, NULL);

  return 0;
}

#else  /* SIGNALS_PEDAL_TO_METAL */

typedef void (*sigaction_proto) (int, siginfo_t *, void *);

static void copy_context (void *dst, void *src)
{
  memcpy (dst, src, CONTEXT_SZ);
}

static void sig_fatal (int signum, siginfo_t *si, void *a3)
{
  ucontext_t *uc;
  struct unicontext ctx;

  ASM_restore_fs_gs ();

  DEREF (si);
  uc = (ucontext_t *) a3;
  sigemptyset (&ctx.set);
  ctx.set.__val[0] = uc->uc_mcontext.oldmask;
  ctx.signum = signum;
  ctx.err = uc->uc_mcontext.gregs[REG_ERR];
  ctx.trapno = uc->uc_mcontext.gregs[REG_TRAPNO];
  ctx.cr2 = uc->uc_mcontext.cr2;
  ctx.regs = &uc->uc_mcontext.gregs;

  signal (SIGABRT, SIG_DFL);
  signal_failed (&ctx);
}

static void sig_handler (int signum, siginfo_t *si, void *a3)
{
  ucontext_t *uc;
  struct unicontext ctx;

  ASM_restore_fs_gs ();

  DEREF (si);
  uc = (ucontext_t *) a3;
  sigemptyset (&ctx.set);
  ctx.set.__val[0] = uc->uc_mcontext.oldmask;
  ctx.signum = signum;
  ctx.err = uc->uc_mcontext.gregs[REG_ERR];
  ctx.trapno = uc->uc_mcontext.gregs[REG_TRAPNO];
  ctx.cr2 = uc->uc_mcontext.cr2;
  ctx.regs = &uc->uc_mcontext.gregs;

  if (SIGFPE == signum) {
    log ("Division by zero\n");
    signal_failed (&ctx);
  }
  else unihandler (&ctx);
}

static void set_handler (int signum, sigaction_proto handler)
{
  struct sigaction sa;

  sigemptyset (&sa.sa_mask);
  sigaddset (&sa.sa_mask, SIGALRM);

  sa.sa_sigaction = handler;
  sa.sa_flags = SA_SIGINFO | SA_RESTART | (sig_stack_size > 0 ? SA_ONSTACK : 0);

  if (-1 == sigaction (signum, &sa, 0)) {
    ERRFail ("sigaction for signal %d failed", signum);
  }
}
#endif

void cpu_x86_outb (CPUX86State *env, int addr, int val)
{
  DEREF (env);
  leverbose ("outb %#x, %#x\n", addr, val);
  EIO_write (addr, 1, val);
  if (0x20 == addr) {
    if (PIC_irq_pending ()) {
      cpu_x86_interrupt (env, CPU_INTERRUPT_HARD);
    }
  }
}

void cpu_x86_outw (CPUX86State *env, int addr, int val)
{
  DEREF (env);
  leverbose ("outw %#x, %#x\n", addr, val);
  if (addr == 0) {
    EIO_write (addr, 1, val & 0xff);
    EIO_write (addr, 1, val >> 8);
  }
  else
    EIO_write (addr, 2, val);
}

void cpu_x86_outl (CPUX86State *env, int addr, int val)
{
  DEREF (env);
  leverbose ("outd %#x, %#x\n", addr, val);
  EIO_write (addr, 4, val);
}

int cpu_x86_inb (CPUX86State *env, int addr)
{
  DEREF (env);
  leverbose ("inb %#x\n", addr);
  return EIO_read (addr, 1);
}

int cpu_x86_inw (CPUX86State *env, int addr)
{
  DEREF (env);
  leverbose ("inw %#x\n", addr);
  return EIO_read (addr, 2);
}

int cpu_x86_inl (CPUX86State *env, int addr)
{
  DEREF (env);
  leverbose ("ind %#x\n", addr);
  return EIO_read (addr, 4);
}

int cpu_x86_get_pic_interrupt (CPUX86State *env)
{
  int nint;
  DEREF (env);

  nint = PIC_run ();
  return nint;
}

int cpu_x86_iret (CPUX86State *env);
int cpu_x86_iret (CPUX86State *env)
{
  DEREF (env);
  return 0;
}

#define EXCP_ICEBP EXCP_DEBUG

#define BOUND(a,cis,oie)(((a)>=(cis))&&((a)<(oie)))

static void Cpu_loop (struct cpu_context *ctx)
{
  int cnt;
  int trapnr;
  char m[X86_EFLAGS_BUF_SIZE];
  ui pc;
  int pm;
  int run;
  CPUX86State *env;
  DEREF (m);

  run = 0;
  loglevel = 0;
  vif = 1;
  env = ctx->env;

  /* env->eflags |= TF_MASK; */
  if (ctx->async) {
    uint8_t *st;

    pm = (env->cr[0] & 1) && (0 == (env->eflags & VM_MASK));
    Fail_if (pm == 1);
    st = (env->segs[R_SS].base + (env->regs[R_ESP] & 0xffff));

    cpu_x86_load_seg (env, R_CS, RED_SEGMENT);
    env->eip = RED_ASYNC_IRET;
    cpu_x86_interrupt (env, CPU_INTERRUPT_HARD);
  }

  while (1) {
    int nint;
    int_handler h;
    intregs r;
    us *st;

    cnt++;
    loop_count++;

    trapnr = cpu_x86_exec (env);
    pm = (env->cr[0] & 1) && (0 == (env->eflags & VM_MASK));
    pc = (ui) (env->segs[R_CS].base + env->eip);

    if (pm && (env->segs[R_SS].flags & DESC_B_MASK))
      st = (us *) (env->segs[R_SS].base + env->regs[R_ESP]);
    else
      st = (us *) (env->segs[R_SS].base + (env->regs[R_ESP] & 0xffff));

    switch (trapnr) {

      case EXCP_ICEBP:
        nint = pc - ((RED_SEGMENT<<4)|RED_INTERRUPT_00);

        if (RED_XMS_FUNC == nint) {

          EAX (r) = QEAX (env);
          EBX (r) = QEBX (env);
          ECX (r) = QECX (env);
          EDX (r) = QEDX (env);
          ESI (r) = QESI (env);
          EDI (r) = QEDI (env);
          DS (r) = QDS (env);
          XMS_func (r);
          QEAX (env) = EAX (r);
          QEBX (env) = EBX (r);
          QECX (env) = ECX (r);
          QEDX (env) = EDX (r);
          QESI (env) = ESI (r);
          QEDI (env) = EDI (r);
          env->regs[R_ESP] += 4;
          env->eip = st[0];
          cpu_x86_load_seg (env, R_CS, st[1]);
          break;
        }

        if (RED_ASYNC_IRET == nint) {
          ldebug ("ASYNC IRET\n");
          in.pic_loop = 0;
          return;
        }

        if (RED_MOUSE_RETF == nint) {
          ldebug ("MOUSE RETF\n");
          return;
        }

        if (! BOUND (nint, 0, 0x100)) {
          BPT;
          continue;
        }

        h = INT_get (nint);
        if (NULL == h) {
          log ("no handler for int %#x\n", nint);
          cpu_x86_dump_state (env, stderr, 0);
          exit (EXIT_FAILURE);
        }

        EAX (r) = QEAX (env);
        EBX (r) = QEBX (env);
        ECX (r) = QECX (env);
        EDX (r) = QEDX (env);
        ESI (r) = QESI (env);
        EDI (r) = QEDI (env);
        EBP (r) = QEBP (env);
        ESP (r) = QESP (env);
        EIP (r) = QEIP (env);
        CS (r) = QCS (env);
        DS (r) = QDS (env);
        ES (r) = QES (env);
        SS (r) = QSS (env);
        FS (r) = QFS (env);
        GS (r) = QGS (env);
        EFL (r) = st[2];
        SETCFL (env, st[2]);

        env->eflags |= IF_MASK;
        outside++;
        /* log ("int %#x AX=%#x\n", nint, AX (r)); */
        /* INT_log_regs (r); */
        h (r);
        outside--;

/*         log ("ret from int %#x ax=%#x efl=%s\n", nint, QEAX (env), */
/*              efltoa (EFL (r), m)); */

        QEAX (env) = EAX (r);
        QEBX (env) = EBX (r);
        QECX (env) = ECX (r);
        QEDX (env) = EDX (r);
        QESI (env) = ESI (r);
        QEDI (env) = EDI (r);
        QEBP (env) = EBP (r);
        QESP (env) = ESP (r);
        cpu_x86_load_seg (env, R_DS, DS (r));
        cpu_x86_load_seg (env, R_ES, ES (r));
        cpu_x86_load_seg (env, R_SS, SS (r));
        cpu_x86_load_seg (env, R_FS, FS (r));
        cpu_x86_load_seg (env, R_GS, GS (r));

        st[2] = EFL (r);

        env->regs[R_ESP] += 6;
        env->eip = st[0];
        cpu_x86_load_seg (env, R_CS, st[1]);
        SETCFL (env, st[2]);
        break;

      default:
        BPT;
    }
  }
}

void SIG_start (regs86 *r)
{
  CPUState *env;
  struct cpu_context ctx;
  struct itimerval iv;
  linfo ("starting\n");

  logfile = fopen ("qemu.log", "w");
  if (NULL == logfile) {
    perror ("fopen qemu.log");
    exit (EXIT_FAILURE);
  }
  setvbuf (logfile, NULL, _IONBF, 0);

  env = global_env;
  env->user_mode_only = 0;
  env->regs[R_EAX] = SIGEAX (r);
  env->regs[R_EBX] = SIGEBX (r);
  env->regs[R_ECX] = SIGECX (r);
  env->regs[R_EDX] = SIGEDX (r);
  env->regs[R_EDI] = SIGEDI (r);
  env->regs[R_ESI] = SIGESI (r);
  env->regs[R_EBP] = SIGEBP (r);
  env->regs[R_ESP] = SIGESP (r);

  env->eip = SIGEIP (r);
  env->eflags |= IF_MASK;

  cpu_x86_load_seg (env, R_CS, SIGCS (r));
  cpu_x86_load_seg (env, R_DS, SIGDS (r));
  cpu_x86_load_seg (env, R_ES, SIGES (r));
  cpu_x86_load_seg (env, R_FS, SIGFS (r));
  cpu_x86_load_seg (env, R_GS, SIGGS (r));
  cpu_x86_load_seg (env, R_SS, SIGSS (r));

  iv.it_interval.tv_sec = 0;
  iv.it_interval.tv_usec = 1000000 / 100;
  iv.it_value.tv_sec = 0;
  iv.it_value.tv_usec = iv.it_interval.tv_usec;

  env->idt.base = (uint8_t *) MEM_REAL (0);
  env->idt.limit = 256 * 4 - 1;

  setitimer (ITIMER_REAL, &iv, 0);

  ctx.env = env;
  ctx.async = 0;
  vif = 1;
  Cpu_loop (&ctx);
}

char *SIG_strdup (const char *s, const char *file,
                  const char *func, int line)
{
  char *ptr;

  r_lock ();
  ptr = strdup (s);
  r_unlock ();

  if (NULL == ptr) {
    ERRFail ("(s=%p, file=`%s', func=`%s', line=%d)",
             s, file, func, line);
  }

  ldebug ("(s=%p, file=`%s', func=`%s', line=%d) = %p\n",
          s, file, func, line, ptr);
  return ptr;
}

void *SIG_malloc (unsigned int size, char *file, const char *func, int line)
{
  void *ptr;

  r_lock ();
  ptr = malloc (size);
  r_unlock ();

  if (NULL == ptr) {
    ERRFail ("(size=%u, file=`%s', func=`%s', line=%d)",
             size, file, func, line);
  }

  ldebug ("(size=%u, file=`%s', func=`%s', line=%d) = %p\n",
          size, file, func, line, ptr);

  return ptr;
}

void SIG_free (void *ptr, char *file, const char *func, int line)
{
  r_lock ();
  free (ptr);
  r_unlock ();

  DEREF (file);
  DEREF (func);
  DEREF (line);

  ldebug ("(ptr=%p, file=`%s', func=`%s', line=%d) = %p\n",
          ptr, file, func, line, ptr);
}

static int no_setter (char *arg)
{
  DEREF (arg);
  return 0;
}

void SIG_pre_init ()
{
  static struct MUP_key keys[] = {
    {"stack.size"        , TAG_IN_INT , {.ii = 16384}},
    {"use.locked.stack"  , TAG_IN_INT , {.ii = 0}},
    {"flat.locked.stack" , TAG_IN_INT , {.ii = 0}},
    {"debugger.command"  , TAG_STRING , {.sp = &debugger_command}},
  };

  static struct MUP_group group =
    {"sig.", LENOFA (keys), keys, NULL};

  static struct MUP_key keys_ign[] =
    {{"cli"               , TAG_SETTER , {.setter = no_setter}},};

  MUP_add_group (&group);
  MUP_add_keys (keys_ign, LENOFA (keys_ign));
}

void SIG_init ()
{
  ub *redir;
  int sig_stack_size;

  log_level = MUP_get_log_level ("sig");

  sig_stack_size = MUP_get_int ("sig.stack.size");

#ifdef SIGNALS_PEDAL_TO_METAL
  if (sig_stack_size <= 0) {
    Fail ("bad sig.stack.size %d, should be positive integer",
	  sig_stack_size);
  }
#endif

  set_stack (sig_stack_size);

  DEREF (is_fault_freq_high);
  debugger_present = is_debugger_present ();

  set_handler (SIGSEGV, sig_handler);
  set_handler (SIGALRM, sig_handler);
  set_handler (SIGUSR1, sig_handler);
  set_handler (SIGTRAP, sig_handler);

  set_handler (SIGFPE, sig_fatal);
  set_handler (SIGBUS, sig_fatal);
  set_handler (SIGILL, sig_fatal);
  set_handler (SIGUSR2, sig_fatal);

  redir = (ub *) RED_FLAT (0);
  redir[RED_ASYNC_IRET] = RED_ICE;
  redir[RED_MOUSE_RETF] = RED_ICE;

  phys_ram_base = (void *) MEM_REAL (0);
  phys_ram_size = MEM_ram_size ();
  phys_ram_fd = MEM_ram_fd ();

  global_env = cpu_init ();
  cpu_single_env = global_env;
  if (NULL == global_env) {
    Fail ("cpu_init failed");
  }

  global_env->cr[0] = 0x0000032;
  cpu_register_physical_memory(0, phys_ram_size, 0);
  cpu_x86_init_mmu (global_env);
}
