#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/syscall.h>
#define LDT_ENTRIES     8192
/* The size of each LDT entry. */
#define LDT_ENTRY_SIZE  8

#include <i86/turtle.h>

#undef SET_CF
#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/int.h>
#include <inc/mem.h>
#include <inc/xdt.h>
#include <inc/sig.h>
#include <inc/eio.h>
#include <inc/mup.h>
#include <inc/log.h>

#include <qemu.h>
#include "vga/vgaemu.h"

extern FILE *turtle_log;
FILE *turtle_log;

static const char * const reg_names[8] = {
  "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi"
};

static const char * const seg_names[6] = {
  "es", "cs", "ss", "ds", "fs", "gs"
};

static CPUReadMemoryFunc *turtle_read_vga_mem8;
static CPUReadMemoryFunc *turtle_read_vga_mem16;
static CPUReadMemoryFunc *turtle_read_vga_mem32;

static CPUWriteMemoryFunc *turtle_write_vga_mem8;
static CPUWriteMemoryFunc *turtle_write_vga_mem16;
static CPUWriteMemoryFunc *turtle_write_vga_mem32;

static int log_level;
#define LOG_PREFIX "X86"

#define TOEMU(env) ((struct X86_emu*)(env))
#define HOT(env) ((struct X86_emu*)(env))->hot=1
#define IS_MEM_VGA(env,a) (((a)>=TOEMU(env)->mem_l)&&((a)<TOEMU(env)->mem_u))

struct X86_emu {
  X86 env;
  CPUX86State *qenv;
  ui mem_l, mem_u;
  int hot;
  int exit;
  int threshold;
  int replay;
};

static int catch_vga;
static struct X86_emu *current;
static struct X86 before_env, after_env;

const uint8_t parity_tab[256] =
  { PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, };

char *X86_efltoa (ui eflags, char m[31])
{
  char *s;

  s = m;
#define C(m,c) *s++ = c ^ ((!!(eflags & m)) << 5)
#define F(n,c) C(n##F,c)
  F(C,'c'); F(P,'p'); F(A,'a'); F(Z,'z'); F(S,'s');
  F(T,'t'); F(I,'i'); F(D,'d'); F(O,'o'); C(VM,'v');
#undef C
#undef F
  snprintf (s, 15,
            " IOPL=%d%s%s%s%s%s", /* 6+1+3*3+2*4) = 24 */
            (eflags >> IOPL_SHIFT) & 3,
            (eflags & NT ? " NT" : ""),
            (eflags & RF ? " RF" : ""),
            (eflags & AC ? " AC" : ""),
            (eflags & VIF ? " VIF" : ""),
            (eflags & VIP ? " VIP" : ""));
  /* 25 */
  /* *s = 0; */
  return m;
}

const char *X86_reg_name (ub nreg)
{
  Fail_if (nreg > 7);
  return reg_names[nreg];
}

const char *X86_seg_name (ub nseg)
{
  Fail_if (nseg > 5);
  return seg_names[nseg];
}

static void log_emu (struct X86_emu *emu)
{
  X86 *env;
  char m[X86_EFLAGS_BUF_SIZE];

  env = &emu->env;

#if 0
  log ("mem_mask = %#x\n"
       "hot = %d\n"
       "exit = %d\n"
       "threshold = %d\n"
       "env = %p\n",
       emu->mem_mask,
       emu->hot,
       emu->exit,
       emu->threshold,
       &emu->env);
#endif

  log ("\neax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
        "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
        "eip=%#010x efl=%#010x %s\n"
        "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
        "fs=%#06x gs=%#06x\n\n",
        env->regs[_REG_EAX].d, env->regs[_REG_EBX].d,
        env->regs[_REG_ECX].d, env->regs[_REG_EDX].d,
        env->regs[_REG_ESI].d, env->regs[_REG_EDI].d,
        env->regs[_REG_EBP].d, env->regs[_REG_ESP].d,
        env->eip, env->eflags,
        X86_efltoa (env->eflags, m),
        env->segs[_SEG_CS].selector,
        env->segs[_SEG_DS].selector,
        env->segs[_SEG_ES].selector,
        env->segs[_SEG_SS].selector,
        env->segs[_SEG_FS].selector,
        env->segs[_SEG_GS].selector
    );
}

uint32_t P__(lea, uint8_t modrm)
{
  int mod;
  int rm;

  mod = (modrm >> 6) & 3;
  rm = modrm & 7;

  /* assert (3 != mod); */

  if (env->a32) {
    uint32_t disp;

    disp = 0;
    if (4 == rm) {
      int index, scale;
      uint8_t sib;

      sib = FETCH_BYTE (env);

      rm = sib & 7;
      index = (sib >> 3) & 7;
      scale = (sib >> 6) & 3;

      if (4 != index)
        disp += READ_DWORD_REG (env, index) << scale;
    }

    if (1 == mod) disp += (int8_t) FETCH_BYTE (env);
    disp += READ_DWORD_REG (env, rm);
    if ((rm | -mod) == 5) {
      disp += FETCH_DWORD (env);
      disp -= READ_DWORD_REG (env, 5);
    }
    else {
      if ((0 == (env->iflags & IF_SEG_OVERRIDE)) && (4 == (rm & 6)))
        env->seg = _SEG_SS;

      if (2 == mod) disp += FETCH_DWORD (env);
    }
    return disp;
  }
  else {
    uint32_t disp = 0;
    switch (mod) {
      case 0:
        if (6 == rm)
          return FETCH_WORD (env);
        break;

      case 2:
        disp = FETCH_WORD (env);
        break;

      case 1:
        disp = (int8_t) FETCH_BYTE (env);
        break;
    }

    if (rm < 6)
      disp += READ_WORD_REG (env, _REG_SI + (rm & 1));

    switch (rm) {
      case 0:                 /* 000 */
      case 1:                 /* 001 */
      case 7:                 /* 111 */
        disp += READ_WORD_REG (env, _REG_BX);
        break;

      case 2:                 /* 010 */
      case 3:                 /* 011 */
      case 6:                 /* 110 */
        disp += READ_WORD_REG (env, _REG_BP);
        if (0 == (env->iflags & IF_SEG_OVERRIDE))
          env->seg = _SEG_SS;
        break;
    }
    return (disp & 0xffff);
  }
}

void P__(write_cr, int rno, uint32_t v)
{
  DEREF (v);
  DEREF (rno);
  turtle_not_implemented (env);
}

void P__(write_dr, int rno, uint32_t v)
{
  DEREF (v);
  DEREF (rno);
  turtle_not_implemented (env);
}

void P0__(not_implemented)
{
  env->iflags |= IF_NOT_IMPLEMENTED;
}

void P0__(confirm)
{
  env->iflags |= IF_CONFIRM;
  memcpy (&before_env, env, sizeof (before_env));
}

void P__(exception, int intno, int err)
{
  if (LOG_DEBUG == log_level) {
    ldebug ("exception %d error %d\n", intno, err);
    log_emu (TOEMU (env));
  }
  env->iflags |= IF_NOT_IMPLEMENTED;
  Fail ("EXCEPTION %#x error %#x", intno, err);
}

void P__(interrupt, int intno)
{
  if (LOG_DEBUG == log_level) {
    ldebug ("interrupt %#x\n", intno);
    log_emu (TOEMU (env));
  }
  env->iflags |= IF_NOT_IMPLEMENTED;
}

static x86exit_e interp (struct X86_emu *emu)
{
  int i, insns;
  X86 *env;

  env = &emu->env;
  i = emu->threshold;

  insns = 0;
  while (i--) {
    int intr;

    if (emu->exit && (0 == (env->iflags & IF_INTR_DISABLED))) {
      ldebug ("exit requested, processed total of %d insns\n", insns);
      return X86_EXIT_REQUESTED;
    }

    env->pc = env->segs[_SEG_CS].base + env->eip;
    env->rep = 0;
    env->a32 = env->o32 = env->c32;
    env->seg = _SEG_DS;

    intr = env->iflags & IF_INHIBIT_INTS;

    env->iflags |= IF_INTR_DISABLED;
    env->iflags &= ~(IF_INHIBIT_INTS | IF_JUMP | IF_SEG_OVERRIDE);
    turtle (env);

    if (env->iflags & IF_NOT_IMPLEMENTED) {
      if (env->segs[_SEG_CS].base[env->eip] != 0xcd) {
        ldebug ("bad at %p %#x %#x %#x\n",
                env->segs[_SEG_CS].base + env->eip,
                env->segs[_SEG_CS].base[env->eip],
                env->segs[_SEG_CS].base[env->eip + 1],
                env->segs[_SEG_CS].base[env->eip + 2]);
      }

      ldebug ("processed total of %d insns\n", insns);
      return X86_EXIT_BAD_INSN;
    }

    if (env->iflags & IF_CONFIRM) {
      ldebug ("confirmation requested, processed total of %d insns\n", insns);
      return X86_EXIT_TEMPORARY;
    }

    if (0 == (env->iflags & IF_JUMP)) {
      env->eip = NEXT_IP (env);
    }

    insns++;

    env->iflags ^= intr;

    if (0 == (env->iflags & IF_INHIBIT_INTS)) {
      env->iflags &= ~IF_INTR_DISABLED;
    }

    if (emu->hot) {
      emu->hot = 0;
      i = emu->threshold;
    }
  }

  ldebug ("threshold processed total of %d insns\n", insns);
  return X86_EXIT_THRESHOLD;
}

uint8_t P__(read_port8, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 1);
}

uint16_t P__(read_port16, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 2);
}

uint32_t P__(read_port32, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 4);
}

void P__(write_port8, uint16_t nport, uint8_t v)
{
  HOT (env);
  EIO_write (nport, 1, v);
}

void P__(write_port16, uint16_t nport, uint16_t v)
{
  HOT (env);
  EIO_write (nport, 2, v);
}

void P__(write_port32, uint16_t nport, uint32_t v)
{
  HOT (env);
  EIO_write (nport, 4, v);
}

#define SIZE 8
#include <inc/mem.inc>
#undef SIZE

#define SIZE 16
#include <inc/mem.inc>
#undef SIZE

#define SIZE 32
#include <inc/mem.inc>
#undef SIZE

/* pushes */
void P__(push16, uint16_t v)
{
  if (IS_STACK_32 (env)) {
    env->regs[_REG_ESP].d -= 2;
    if (0 == TOEMU (env)->replay)
      *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d) = v;
  }
  else {
    env->regs[_REG_SP].w -= 2;
    if (0 == TOEMU (env)->replay)
      *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w) = v;
  }
}

void P__(push32, uint32_t v)
{
  if (IS_STACK_32 (env)) {
    env->regs[_REG_ESP].d -= 4;
    if (0 == TOEMU (env)->replay)
      *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d) = v;
  }
  else {
    env->regs[_REG_SP].w -= 4;
    if (0 == TOEMU (env)->replay)
      *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w) = v;
  }
}

uint16_t P0__(pop16)
{
  uint16_t v;

  v = 0;
  if (IS_STACK_32 (env)) {
    if (0 == TOEMU (env)->replay)
      v = *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d);
    env->regs[_REG_ESP].d += 2;
  }
  else {
    if (0 == TOEMU (env)->replay)
      v = *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w);
    env->regs[_REG_SP].w += 2;
  }

  return v;
}

uint32_t P0__(pop32)
{
  uint32_t v;

  v = 0;
  if (IS_STACK_32 (env)) {
    if (0 == TOEMU (env)->replay)
      v = *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d);
    env->regs[_REG_ESP].d += 4;
  }
  else {
    if (0 == TOEMU (env)->replay)
      v = *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w);
    env->regs[_REG_SP].w += 4;
  }
  return v;
}

void P0__(cli)
{
#ifdef IF_MANAGEMENT
  uint32_t f;

  f = env->eflags;

  if (0 == PE (env)) {
    env->eflags = f & ~ IF;
  }
  else {
    int cpl, iopl;

    cpl = CPL (env);
    iopl = (f & IOPL_MASK) >> IOPL_SHIFT;

    if (0 == (f & VM)) {
      if (iopl >= cpl) {
        env->eflags = f & ~ IF;
      }
      else {
        if ((iopl < cpl) && (cpl < 3) && PVI (env)) {
          env->eflags = f & ~ VIF;
        }
        else {
          turtle_exception (env, 13, 0);
        }
      }
    }
    else {
      if (3 == iopl) {
        env->eflags = f & ~ IF;
      }
      else {
        if ((iopl < 3) && VME (env)) {
          env->eflags = f & ~ VIF;
        }
        else {
          turtle_exception (env, 13, 0);
        }
      }
    }
  }
#else
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(sti)
{
#ifdef IF_MANAGEMENT
  uint32_t f;

  f = env->eflags;

  if (0 == PE (env)) {
    env->eflags = f | IF;
  }
  else {
    int cpl;
    int iopl;

    cpl = CPL (env);
    iopl = (f & IOPL_MASK) >> IOPL_SHIFT;

    if (0 == (f & VM)) {
      if (iopl >= cpl) {
        env->eflags = f | IF;
        INHIBIT_INTERRUPTS (env);
      }
      else {
        if ((iopl < cpl) && (3 == cpl) && (0 == (f & VIP))) {
          env->eflags = f | VIF;
        }
        else {
          env->eflags = f;
          EXCEPTION (env, 13, 0);
        }
      }
    }
    else {
      if (3 == iopl) {
        env->eflags = f | IF;
        INHIBIT_INTERRUPTS (env);
      }
      else {
        if ((iopl < 3) && (0 == (f & VIP)) && VME (env)) {
          env->eflags = f | VIF;
        }
        else {
          EXCEPTION (env, 13, 0);
        }
      }
    }
  }
#else
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

static void pushf (X86 *env, int d)
{
#ifdef IF_MANAGEMENT
  int pe, vm, iopl;

  pe = PE (env);
  vm = 0 != (env->eflags & VM);
  iopl = (env->eflags & IOPL_MASK) >> IOPL_SHIFT;

  if ((0 == pe) || (pe && ((0 == vm) || (vm && (3 == iopl))))) {
    if (d)
      turtle_push32 (env, env->eflags & 0xfcffff);
    else
      turtle_push16 (env, env->eflags);
  }
  else {
    turtle_exception (env, 13, 0);
  }
#else
  DEREF (env);
  DEREF (d);
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(pushf16)
{
  pushf (env, 0);
}

void P0__(pushf32)
{
  pushf (env, 1);
}

static void popf (X86 *env, int d)
{
#ifdef IF_MANAGEMENT
  uint32_t rF, wF;

  rF = env->eflags;
  wF = rF;
  if (0 == (rF & VM)) {
    int cpl;

    cpl = CPL (env);
    if (d)
      wF = turtle_pop32 (env);
    else
      wF = turtle_pop16 (env);

    SANITIZE_EFLAGS (wF);

    if (0 == cpl) {
      wF &= ~ (VIF | VIP | VM);
      wF |= rF & VM;
    }
    else {
      wF &= ~ (VIF | VIP | VM | IOPL_MASK);
      wF |= rF & (VM | (3 << IOPL_SHIFT));
    }
  }
  else {
    if (3 == ((rF & IOPL_MASK) >> IOPL_SHIFT)) {
      if (d)
        wF = turtle_pop32 (env);
      else
        wF = turtle_pop16 (env);
      SANITIZE_EFLAGS (wF);
      wF &= ~ (VIF | VIP | VM | RF | IOPL_MASK);
      wF |= rF & (VIF | VIP | VM | RF | IOPL_MASK);
    }
    else {
      EXCEPTION (env, 13, 0);
    }
  }
  if ((rF & IF) && (0 == (wF & IF))) {
    ldebug ("popf disabled interrupts\n");
  }
  env->eflags = wF;
#else
  DEREF (env);
  DEREF (d);
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(popf16)
{
  popf (env, 0);
}

void P0__(popf32)
{
  popf (env, 1);
}

static void read_ldt (int index, void *buf)
{
  int count;
  char map[LDT_ENTRIES * LDT_ENTRY_SIZE];

  count = syscall (SYS_modify_ldt, 0, &map, 8); /* sizeof (map)); */

  if (8 != count) {
    ERRFail ("modify_ldt(read) %d failed, count %d", index, count);
  }

  memcpy (buf, map + LDT_ENTRY_SIZE * index, LDT_ENTRY_SIZE);
}

void P__(load_segment, int seg, uint32_t selector)
{
  struct X86_emu *emu;
  unsigned int ar;
  int limit;
  unsigned long base;

  emu = TOEMU (env);

  if (NULL == emu->qenv) {
    if (4 == (selector & 4)) {
      static int done;
      static ul d;
      if (0 == done) {
        read_ldt (selector >> 3, &d);
        done = 1;
      }
      limit = XDT_get_descriptor_limit (&d);
      base = XDT_get_descriptor_base (&d);
      ar = (uint32_t) (d >> 32U);
    }
    else {
      base = 0;
      if (0 == selector) {
        ar = 0;
        base = -1;
      }
      else if (ASM_cs == selector)
        ar = 0xcffa00;
      else if (selector == ASM_ds)
        ar = 0xcff300;
      else {
        turtle_not_implemented (env);
        return;
      }
      limit = -1;
    }
    env->segs[seg].selector = (selector & 0xffff);
    env->segs[seg].base = (uint8_t *) base;
    env->segs[seg].limit = limit;
    env->segs[seg].flags = ar;

    if (_SEG_CS == seg) env->c32 = (ar >> 22) & 1;
    if (_SEG_SS == seg) env->s32 = (ar >> 22) & 1;
  }
  else {
    if (PE (env) && (0 == (env->eflags & VM))) {
      ul d;
      /* printf ("loading seg %#x selector %#x\n", seg, selector); */

      if (4 == (selector & 4)) {
        d = *(ul *) (emu->qenv->ldt.base + (selector >> 3));
      }
      else {
        d = *(ul *) (emu->qenv->gdt.base + (selector >> 3));
      }

      limit = XDT_get_descriptor_limit (&d);
      base = XDT_get_descriptor_base (&d);
      ar = (uint32_t) (d >> 32U);

      env->segs[seg].selector = (selector & 0xffff);
      env->segs[seg].base = (uint8_t *) base;
      env->segs[seg].limit = limit;
      env->segs[seg].flags = ar;

      if (_SEG_CS == seg) env->c32 = (ar >> 22) & 1;
      if (_SEG_SS == seg) env->s32 = (ar >> 22) & 1;
    }
    else {
      env->segs[seg].selector = (selector & 0xffff);
      env->segs[seg].base = (uint8_t *) ((selector & 0xffff) << 4);
      env->segs[seg].limit = 0xffff;
      env->segs[seg].flags = 0;
      if (_SEG_CS == seg) env->c32 = 0;
      if (_SEG_SS == seg) env->s32 = 0;
    }
  }
}

x86exit_e X86_confirm32 (regs86 *r, int threshold)
{
  int i;
  ui mask;
  X86 *env;
  int diff;

  env = &after_env;
  diff = 0;

  for (i = 0; i < 8; i++) {
    uint32_t hw, sw;

    hw = *SIG_gregp (r, i);
    sw = READ_DWORD_REG (env, i);
    if (hw != sw) {
      log ("%s %#010x %#010x\n", reg_names[i], hw, sw);
      diff = 1;
    }
  }

  for (i = 0; i < 6; i++) {
    uint16_t hw, sw;

    hw = *SIG_sregp (r, i);
    sw = READ_SEGMENT_REG (env, i);
    if (hw != sw) {
      log ("%s %#06x %#06x\n", seg_names[i], hw, sw);
    }
    diff = 1;
  }

  if ((uint32_t) SIGEIP (r) != env->eip) {
    log ("EIP %p %p\n",
         env->segs[_SEG_CS].base + SIGEIP (r),
         env->segs[_SEG_CS].base + env->eip);
    diff = 1;
  }

  mask = IF|TF|RF|(3<<IOPL_SHIFT);
  if ((SIGEFL (r) & ~mask) != (env->eflags & ~mask)) {
    char m1[X86_EFLAGS_BUF_SIZE];
    char m2[X86_EFLAGS_BUF_SIZE];

    X86_efltoa (SIGEFL (r), m1);
    X86_efltoa (env->eflags, m2);

    log ("%#10x EFL %#7x %s %#7x %s\n",
         (int) SIGEIP (r), (int) SIGEFL (r), m1, env->eflags, m2);
    diff = 0;
  }
  if (diff) {
    log ("DIFFERENT\n");
    abort ();
  }
  return X86_run32 (r, threshold);
}


x86exit_e X86_confirm (CPUState *qenv, int threshold)
{
  int i;
  ui mask;
  X86 *env;
  int diff;

  env = &after_env;
  diff = 0;

  for (i = 0; i < 8; i++) {
    uint32_t hw, sw;

    hw = qenv->regs[i];
    sw = READ_DWORD_REG (env, i);
    if (hw != sw) {
      log ("%s %#010x %#010x\n", reg_names[i], hw, sw);
      diff = 1;
    }
  }

  for (i = 0; i < 6; i++) {
    uint16_t hw, sw;

    hw = qenv->segs[i].selector;
    sw = READ_SEGMENT_REG (env, i);
    if (hw != sw) {
      log ("%s %#06x %#06x\n", seg_names[i], hw, sw);
    }
    diff = 1;
  }

  if (qenv->eip != env->eip) {
    log ("EIP %p %p\n",
         env->segs[_SEG_CS].base + qenv->eip,
         env->segs[_SEG_CS].base + env->eip);
    diff = 1;
  }

  mask = IF|TF|RF|(3<<IOPL_SHIFT);
  if ((qenv->eflags & ~mask) != (env->eflags & ~mask)) {
    char m1[X86_EFLAGS_BUF_SIZE];
    char m2[X86_EFLAGS_BUF_SIZE];

    X86_efltoa (qenv->eflags, m1);
    X86_efltoa (env->eflags, m2);

    log ("%#10x EFL %#7x %s %#7x %s\n",
         qenv->eflags, qenv->eflags, m1, env->eflags, m2);
    diff = 0;
  }

  if (diff) {
    log ("DIFFERENT\n");
    abort ();
  }

  return X86_run (qenv, threshold);
}

#define IO_MEM_NB_ENTRIES 100
CPUWriteMemoryFunc *io_mem_write[IO_MEM_NB_ENTRIES][4];
CPUReadMemoryFunc *io_mem_read[IO_MEM_NB_ENTRIES][4];

x86exit_e X86_run32 (regs86 *r, int threshold)
{
  /* FILE *save; */
  X86 *env;
  int was_enabled;
  x86exit_e exit;
  struct X86_emu emu;

  memset (&emu, 0, sizeof (emu));

  /* save = turtle_log; */
  /* turtle_log = NULL; */

  turtle_write_vga_mem8 = io_mem_write[2][0];
  turtle_write_vga_mem16 = io_mem_write[2][1];
  turtle_write_vga_mem32 = io_mem_write[2][2];

  turtle_read_vga_mem8 = io_mem_read[2][0];
  turtle_read_vga_mem16 = io_mem_read[2][1];
  turtle_read_vga_mem32 = io_mem_read[2][2];

  env = &emu.env;
  env->cr[0] = 1;

  was_enabled = INT_is_enabled ();
  env->eflags = SIGEFL (r) | (3 << IOPL_SHIFT);
  if (was_enabled)
    env->eflags |= IF_MASK;
  else
    env->eflags &= ~IF_MASK;

  env->eip = SIGEIP (r);
  env->iflags = 0;

  WRITE_DWORD_REG (env, _REG_EAX, SIGEAX (r));
  WRITE_DWORD_REG (env, _REG_EBX, SIGEBX (r));
  WRITE_DWORD_REG (env, _REG_ECX, SIGECX (r));
  WRITE_DWORD_REG (env, _REG_EDX, SIGEDX (r));
  WRITE_DWORD_REG (env, _REG_ESI, SIGESI (r));
  WRITE_DWORD_REG (env, _REG_EDI, SIGEDI (r));
  WRITE_DWORD_REG (env, _REG_EBP, SIGEBP (r));
  WRITE_DWORD_REG (env, _REG_ESP, SIGESP (r));
  WRITE_SEGMENT_REG (env, _SEG_CS, SIGCS (r));
  WRITE_SEGMENT_REG (env, _SEG_DS, SIGDS (r));
  WRITE_SEGMENT_REG (env, _SEG_ES, SIGES (r));
  WRITE_SEGMENT_REG (env, _SEG_SS, SIGSS (r));
  WRITE_SEGMENT_REG (env, _SEG_FS, SIGFS (r));
  WRITE_SEGMENT_REG (env, _SEG_GS, SIGGS (r));

  /* turtle_log = save; */

  emu.threshold = threshold;
  emu.hot = 0;
  emu.exit = 0;
  emu.replay = 0;

  if (catch_vga) {
    emu.mem_l = 0xa0000;
    emu.mem_u = 0xc0000;
  }

  ldebug ("interpreting at %p\n", env->segs[_SEG_CS].base + env->eip);

  current = &emu;
  exit = interp (&emu);
  current = NULL;

  if (X86_EXIT_TEMPORARY == exit) {
    memcpy (&after_env, env, sizeof (after_env));
    after_env.eip = NEXT_IP ((&after_env));
    env = &before_env;
  }

  SIGEAX (r) = READ_DWORD_REG (env, _REG_EAX);
  SIGEBX (r) = READ_DWORD_REG (env, _REG_EBX);
  SIGECX (r) = READ_DWORD_REG (env, _REG_ECX);
  SIGEDX (r) = READ_DWORD_REG (env, _REG_EDX);
  SIGESI (r) = READ_DWORD_REG (env, _REG_ESI);
  SIGEDI (r) = READ_DWORD_REG (env, _REG_EDI);
  SIGEBP (r) = READ_DWORD_REG (env, _REG_EBP);
  SIGESP (r) = READ_DWORD_REG (env, _REG_ESP);
  SIGCS (r) = READ_SEGMENT_REG (env, _SEG_CS);
  SIGDS (r) = READ_SEGMENT_REG (env, _SEG_DS);
  SIGES (r) = READ_SEGMENT_REG (env, _SEG_ES);
  SIGSS (r) = READ_SEGMENT_REG (env, _SEG_SS);
  SIGFS (r) = READ_SEGMENT_REG (env, _SEG_FS);
  SIGGS (r) = READ_SEGMENT_REG (env, _SEG_GS);
  SIGEIP (r) = env->eip;
  SIGEFL (r) = env->eflags;

  if (env->eflags & IF) {
    if (0 == was_enabled)
      abort ();
    INT_enable ();
  }
  else {
    if (was_enabled)
      abort ();
    INT_disable ();
  }

  /* turtle_log = save; */
  return exit;
}

x86exit_e X86_run (CPUState *qenv, int threshold)
{
  FILE *save;
  X86 *env;
  x86exit_e exit;
  struct X86_emu emu;

  memset (&emu, 0, sizeof (emu));

  save = turtle_log;
  turtle_log = NULL;

  emu.qenv = qenv;
  env = &emu.env;
  env->cr[0] = qenv->cr[0];

  env->eflags = qenv->eflags;

  env->eip = qenv->eip;
  env->iflags = 0;

  WRITE_DWORD_REG (env, _REG_EAX, qenv->regs[R_EAX]);
  WRITE_DWORD_REG (env, _REG_EBX, qenv->regs[R_EBX]);
  WRITE_DWORD_REG (env, _REG_ECX, qenv->regs[R_ECX]);
  WRITE_DWORD_REG (env, _REG_EDX, qenv->regs[R_EDX]);
  WRITE_DWORD_REG (env, _REG_ESI, qenv->regs[R_ESI]);
  WRITE_DWORD_REG (env, _REG_EDI, qenv->regs[R_EDI]);
  WRITE_DWORD_REG (env, _REG_EBP, qenv->regs[R_EBP]);
  WRITE_DWORD_REG (env, _REG_ESP, qenv->regs[R_ESP]);
  WRITE_SEGMENT_REG (env, _SEG_CS, qenv->segs[R_CS].selector);
  WRITE_SEGMENT_REG (env, _SEG_DS, qenv->segs[R_DS].selector);
  WRITE_SEGMENT_REG (env, _SEG_ES, qenv->segs[R_ES].selector);
  WRITE_SEGMENT_REG (env, _SEG_SS, qenv->segs[R_SS].selector);
  WRITE_SEGMENT_REG (env, _SEG_FS, qenv->segs[R_FS].selector);
  WRITE_SEGMENT_REG (env, _SEG_GS, qenv->segs[R_GS].selector);

  emu.threshold = threshold;
  emu.hot = 0;
  emu.exit = 0;
  emu.replay = 0;

  if (catch_vga) {
    emu.mem_l = 0xa0000;
    emu.mem_u = 0xc0000;
  }

  ldebug ("interpreting at %p\n", env->segs[_SEG_CS].base + env->eip);

  current = &emu;
  exit = interp (&emu);
  current = NULL;

  if (X86_EXIT_TEMPORARY == exit) {
    memcpy (&after_env, env, sizeof (after_env));
    after_env.eip = NEXT_IP ((&after_env));
    env = &before_env;
  }

  qenv->eflags = env->eflags;
  qenv->regs[R_EAX] = READ_DWORD_REG (env, _REG_EAX);
  qenv->regs[R_EBX] = READ_DWORD_REG (env, _REG_EBX);
  qenv->regs[R_ECX] = READ_DWORD_REG (env, _REG_ECX);
  qenv->regs[R_EDX] = READ_DWORD_REG (env, _REG_EDX);
  qenv->regs[R_ESI] = READ_DWORD_REG (env, _REG_ESI);
  qenv->regs[R_EDI] = READ_DWORD_REG (env, _REG_EDI);
  qenv->regs[R_EBP] = READ_DWORD_REG (env, _REG_EBP);
  qenv->regs[R_ESP] = READ_DWORD_REG (env, _REG_ESP);
  cpu_x86_load_seg (qenv, R_CS, READ_SEGMENT_REG (env, _SEG_CS));
  cpu_x86_load_seg (qenv, R_ES, READ_SEGMENT_REG (env, _SEG_ES));
  cpu_x86_load_seg (qenv, R_SS, READ_SEGMENT_REG (env, _SEG_SS));
  cpu_x86_load_seg (qenv, R_DS, READ_SEGMENT_REG (env, _SEG_DS));
  cpu_x86_load_seg (qenv, R_FS, READ_SEGMENT_REG (env, _SEG_FS));
  cpu_x86_load_seg (qenv, R_GS, READ_SEGMENT_REG (env, _SEG_GS));
  qenv->eip = env->eip;

  turtle_log = save;
  return exit;
}

void X86_stop ()
{
  if (current)
    current->exit = 1;
  else
    lwarn ("no virtual X86 context\n");
}

void X86_dump_current ()
{
  if (current) {
    log ("virtual X86 context\n");
    log_emu (current);
  }
}

void X86_catch_VGA (int do_catch)
{
  catch_vga = do_catch;
}

void X86_pre_init ()
{
  static char *turtle_log_name;

  struct MUP_key keys[] =
    {{"turtle.log", TAG_STRING, {.sp = &turtle_log_name}}};

  MUP_add_keys (keys, LENOFA (keys));
}

void X86_init ()
{
  char *turtle_log_name;

  log_level = MUP_get_log_level ("x86");
  turtle_log_name = MUP_get_string ("turtle.log");

  if (turtle_log_name) {
    turtle_log = fopen (turtle_log_name, "w");
    if (NULL == turtle_log) {
      ERRFail ("failed to open `%s'", turtle_log_name);
    }

    setbuf (turtle_log, NULL);
  }
}
