#include <inc/cmos.h>
#include <inc/eio.h>
#include <inc/rtc.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "CMOS"
static int log_level;

static int mi;
static ub mem[128];

static void write_70 (int nport, int size, ui val)
{
  DEREF (nport);
  DEREF (size);

  EIO_LOG_WRITE;

  mi = val;
}

static ui pirc_to_hz_tab[] = {
  0, 256, 128, 8192, 4096, 2048, 1024,
  512, 256, 128, 64, 32, 16, 8, 4, 2
};

static void write_71 (int nport, int size, ui val)
{
  char *msg;

  msg = (char *) -1;
  DEREF (nport);
  DEREF (size);

  EIO_LOG_WRITE;

  switch (mi) {
    case 0xa:
    {
      int pirc;

      if (val & 0x80) {
        lwarn ("UIP is set this is moronic to say the least\n");
      }

      val &= ~0x80;

      /* pcb = (val >> 4) & 7; */
      pirc = val & 15;

      RTC_set_rate (pirc_to_hz_tab[pirc]);
      mem[mi] = val;
      break;
    }

    case 0xb:
      if (val & 0x80) {
        msg = "set mode not supported\n";
        goto error;
      }

      if (val & 0x20) {
        msg = "AIE not supported\n";
        goto error;
      }

      if (val & 0x10) {
        msg = "UIE not supported\n";
        goto error;
      }

      if (val & 0x8) {
        msg = "SQWE? get real\n";
        goto error;
      }

      if (0 == (val & 0x4)) {
        msg = "sorry no BCD today\n";
        goto error;
      }

      if (val & 0x40)
        RTC_enable ();
      else
        RTC_disable ();

      mem[mi] = val;
      break;


    default:
      msg = "what?";
      goto error;
  }

  return;

 error:
  Fail ("(nport=%#x, size=%d, val=%#x) %s",
        nport, size, val, msg);
}

static ui read_71 (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  EIO_LOG_READ;
  return mem[mi];
}

static ui read_70 (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  lwarn ("eh?");
  return 0;
}

void CMOS_init ()
{
  mem[10] = 9;
  mem[11] = 64 | 4;

  mem[0x15] = 0x80;
  mem[0x16] = 0x02;
  mem[0x17] = 0x00;
  mem[0x18] = 0x40;

  log_level = MUP_get_log_level ("cmos");
  EIO_register_write_handler (0x70, EIO_MASK_B, write_70);
  EIO_register_write_handler (0x71, EIO_MASK_B, write_71);
  EIO_register_read_handler (0x71, EIO_MASK_B, read_71);
  EIO_register_read_handler (0x70, EIO_MASK_B, read_70);
}
