/* all hail lazyness and stupidity */
#include <string.h>
#include <stdlib.h>

#include <inc/mem.h>
#include <inc/dmm.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "DMM"
static int log_level;

static int total_npages;
static ub *pages;
static ui lar_size, lar_base;
static ui base;
static int page_size = 0x10000;
static int topdown;
static int alternate;

static inline ui top (ui i)
{
  return (i + (page_size - 1)) / page_size;
}

static void free_pages (int fbit, int npages)
{
  int i;
  ui lin, size;

  lin = base + fbit * page_size;
  size = npages * page_size;

  if (lin + size == lar_base) {
    lar_base = lin;
    lar_size += size;
  }
  else if (lar_base + lar_size == lin) {
    lar_size += size;
  }
  else if (size > lar_size) {
    lar_base = lin;
    lar_size = size;
  }

  for (i = 0; i < npages; i++)
    pages[(fbit + i) / 8] &= ~(1 << ((i + fbit) % 8));
}

static int alloc_pages (int npages)
{
  ui lin;
  int i, fbit;
  int region = 0;

  fbit = -1;
  for (i = 0; i < total_npages; i++) {
    int j, b, k;

    if (topdown)
      k = total_npages - 1 - i;
    else
      k = i;

    j = k / 8;
    b = 1 << (k % 8);
    if (pages[j] & b) {
      region = 0;
    }
    else {
      if (0 == region) {
        fbit = k;
      }
      if (++region == npages)
        break;
    }
  }

  if (region < npages)
    return -1;

  if (topdown)
    fbit -= npages - 1;

  if (alternate)
    topdown = !topdown;

  ldebug ("fbit = %d\n", fbit);
  for (i = 0; i < npages; i++) {
    int byte, bit;

    byte = (i + fbit) / 8;
    bit = 1 << ((fbit + i) % 8);
    pages[byte] |= bit;
  }

  lin = fbit * page_size + base;

  if ((lin >= lar_base) && (lin < (lar_base + lar_size))) {
    int l_size, r_size;

    l_size = lin - lar_base;
    r_size = (lar_base + lar_size) - (lin + npages * page_size);
    if (l_size > r_size) {
      lar_size = l_size;
    }
    else {
      lar_base = lin + npages * page_size;
      lar_size = r_size;
    }
  }

  return fbit * page_size;
}

void *DMM_alloc (int size)
{
  int res;
  int npages;

  npages = top (size);
  res = alloc_pages (npages);

  if (-1 == res) {
    lwarn ("not enough memory to allocate %d bytes of memory\n", size);
    return NULL;
  }
  linfo ("size=%d res=%d ret=%p\n", size, res, (void *) (base + res));
  memset ((void *) (base + res), 0, npages * page_size);
  return (void *) (base + res);
}

void *DMM_realloc (void *ptr, int old_size, int size)
{
  int res;
  void *ret;
  int npages;
  int old_size_al;

  linfo ("ptr=%p old_size=%d size=%d\n", ptr, old_size, size);

  npages = top (size);
  res = alloc_pages (npages);

  if (-1 == res) {
    lwarn ("not enough memory to resize block %p from %d to %d\n",
           ptr, old_size, size);
    return NULL;
  }
  ret = (void *) (base + res);

  memcpy (ret, ptr, old_size > size ? size : old_size);

  old_size_al = top (old_size);
  free_pages (((ui) ptr - base) / page_size, old_size_al);

  return ret;
}

void DMM_free (void *ptr, int size)
{
  linfo ("(ptr=%p, size=%d)\n", ptr, size);
  size = top (size);
  free_pages (((ui) ptr - base) / page_size, size);
}

ui DMM_largest ()
{
  return lar_size;
}

void DMM_pre_init ()
{
  static struct MUP_key keys[] = {
    {"page.size"          , TAG_INT , {.ip = &page_size}},
    {"topdown"            , TAG_INT , {.ip = &topdown}},
    {"alternate"          , TAG_INT , {.ip = &alternate}},
  };

  static struct MUP_group group =
    {"dmm.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void DMM_init ()
{
  int size, nbytes, total;

  log_level = MUP_get_log_level ("dmm");

  base      = 0x100000;
  total     = MEM_ram_size ();

  size = total - base;
  base = MEM_REAL (base);

  total_npages = (size + page_size - 1) / page_size;
  nbytes = (total_npages + 7) / 8;

  linfo ("size=%d total_npages=%d nbytes=%d page_size=%d\n",
         size, total_npages, nbytes, page_size);

  pages = calloc (nbytes, 1);
  if (NULL == pages) {
    ERRFail ("failed to allocate %d bytes for page map", nbytes);
  }

  lar_base = base;
  lar_size = size;
  MEM_MEM_SIZE = (total / 1024) > 0x8000 ? 0x8000 : (total / 1024);
}
