#include <inc/eio.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "EIO"
static int log_level;

static struct {
  int srmask;
  int swmask;
  EIO_read_handler read_handler;
  EIO_write_handler write_handler;
} handlers[0x400];

static char *bit_string (us val)
{
  int i;
  static char buf[18];

  for (i = 0; i < 16; i++) {
    buf[i + (i > 7)] = (0 != (val & (1 << (15 - i)))) + '0';
  }
  buf[8] = ' ';
  buf[17] = 0;
  return buf;
}

static ui default_reader (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  lwarn ("(nport=%#x, size=%d)\n",
         nport, size);

  return ~0U;
}

static void default_writer (int nport, int size, ui val)
{
  DEREF (nport);
  DEREF (size);
  DEREF (val);

  lwarn ("(nport=%#x, size=%d, val=%#6x %s)\n",
         nport, size, val, bit_string (val));
}

void EIO_register_read_handler (int nport, int smask,
                                EIO_read_handler read_handler)
{
  handlers[nport].read_handler = read_handler;
  handlers[nport].srmask = smask;
}

void EIO_register_write_handler (int nport, int smask,
                                 EIO_write_handler write_handler)
{
  handlers[nport].write_handler = write_handler;
  handlers[nport].swmask = smask;
}

ui EIO_read (int nport, int size)
{
  if (nport > LENOFA (handlers)) {
    if (0xff81 == nport)        /* Bochs port */
      return ~0U;

    lwarn ("(nport=%#x, size=%d) port to big\n",
           nport, size);

    return ~0U;
  }

  if (handlers[nport].srmask & size) {
    ui val;

    val = handlers[nport].read_handler (nport, size);
    ldebug ("(nport=%#x, size=%d) = %#x\n", nport, size, val);
    return val;
  }
  else {
    Fail ("(nport=%#x, size=%d) no handler installed for this size",
           nport, size);
  }
}

void EIO_write (int nport, int size, ui val)
{
  ldebug ("(nport=%#x, size=%d, val=%#x)\n",
          nport, size, val);

  if (nport > LENOFA (handlers)) {
    if ((0xff80 == nport) || (0xff81 == nport) || (0x500 == nport))
      /* Bochs ports */
      return;

    lwarn ("(nport=%#x, size=%d, val=%#x) port to big\n",
           nport, size, val);
    return;
  }

  if (handlers[nport].swmask & size) {
    handlers[nport].write_handler (nport, size, val);
  }
  else {
    Fail ("(nport=%#x, size=%d, val=%#x) no handler installed for this size",
          nport, size, val);
  }
}

void EIO_init ()
{
  int i;

  DEREF (bit_string);
  log_level = MUP_get_log_level ("eio");

  for (i = 0; i < LENOFA (handlers); i++) {
    EIO_register_read_handler (i, EIO_MASK_BWD, default_reader);
    EIO_register_write_handler (i, EIO_MASK_BWD, default_writer);
  }
}
