#include <stdlib.h>
#include <string.h>

#include <inc/sel.h>
#include <inc/int.h>
#include <inc/mem.h>
#include <inc/vid.h>
#include <inc/vga.h>
#include <inc/vbe.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "I10"
static int log_level = -2;

static void vesa (INT_ARGS)
{
  switch (AL (r)) {
    case 0x00:
    {
      char *vi;

      vi = (char *) MEM_REAL (SEGLIN (ES (r), DI (r)));

      VBE_get_vesa_info (vi);
      break;
    }

    case 0x01:
    {
      char *mi;

      mi = (char *) MEM_REAL (SEGLIN (ES (r), DI (r)));

      if (0xffff == CX (r)) {
        SET_AX (r, 0x034f);
        lwarn ("request for strange video mode %#x\n", CX (r));
        return;
      }

      if (0 == VBE_get_mode_info (CX (r), mi)) {
        lwarn ("get mode info for %#x failed\n", CX (r));
        goto error_ok;
      }

        /* goto error_ok; */
      break;
    }

    case 0x02:
      if (0 == VBE_set_mode (BX (r)))
        goto error;
      break;

    case 0x03:
      SET_BX (r, 3 | (1 << 14));
      break;

    case 0x05:
      if (BL (r) > 1)
        goto error;

      switch (BH (r)) {
        case 0:
          VBE_set_page (BL(r), DX (r));
          break;

        case 1:
          SET_DX (r, VBE_get_page (BL (r)));
          break;
      }

      break;

    case 0x06:
      /* SET_AX (r, 0x0100); */
      SET_BX (r, VBE_get_bpsl ());
      SET_CX (r, VBE_get_psl ());
      SET_DX (r, VBE_get_nsl ());
      /* return; */
      break;

/* VESA SuperVGA BIOS v1.1+ - GET/SET DISPLAY START */
    case 0x07:                  /* mostly used for vertical retrace */
      if (1 == BL (r)) {
        SET_BX (r, 0);
        SET_DX (r, 0);
      }
      linfo ("U VESA SuperVGA BIOS v1.1+ - GET/SET DISPLAY START %#x\n"
             "BX = %#x leftmost CX = %d, first displayed scan line DX = %d\n",
             AX (r), BX (r), CX (r), DX (r));
      break;

    case 0x09:
      lwarn ("U VESA SuperVGA BIOS v2.0+ - GET/SET PALETTE ENTRIES %#x\n",
             AX (r));
      goto error_ok;

    case 0x0a:
    {
      ui addr;
      ui size;
      void *v_addr;

      if (0 != BL (r))
        goto error;

      VBE_get_pm_interface (&v_addr, &size);
      addr = MEM_USER ((ui) v_addr);
      ES (r) = (addr >> 4) & 0xffff;
      SET_DI (r, addr - (ES (r) << 4));
      SET_CX (r, size);
      SET_SI (r, 0);
      break;
    }

    default:
      UNHANDLED_INT (r);
  }

  SET_AX (r, 0x4f);
  return;

 error:
  UNHANDLED_INT (r);

 error_ok:
  SET_AX (r, 0x014f);
}

void I10 (INT_ARGS)
{
  if (-2 == log_level)
    log_level = MUP_get_log_level ("i10");

  if (log_level < LOG_DEBUG)
    linfo ("ax = %#06x\n", AX (r));
  else {
    log ("ax = %#06x ", AX (r));
    INT_log_regs (r);
  }

  switch (AH (r)) {
    case 0x00:
      VGA_set_mode (AL (r));
      break;

    case 0x01:                  /* cursor shape */
      lwarn ("U VIDEO - SET TEXT-MODE CURSOR SHAPE %#x\n", AH (r));
      break;

    case 0x02:
      VGA_set_cursor_position (BH (r), DH (r), DL (r));
      break;

    case 0x03:
      SET_CX (r, 0);
      SET_DH (r, MEM_VID_CURSOR_POSITIONS[1]);
      SET_DL (r, MEM_VID_CURSOR_POSITIONS[0]);
      break;

    case 0x05:                  /* set active display page */
      lwarn ("U VIDEO -  SELECT ACTIVE DISPLAY PAGE %#x\n", AX (r));
      break;

    case 0x06:                  /* scroll up */
      lwarn ("U VIDEO - SCROLL UP WINDOW %#x\n", AX (r));
      break;

    case 0x08:
      if (BH (r) != 0) {
        Fail ("VIDEO - READ CHARACTER AND ATTRIBUTE AT CURSOR POSITION %#x %#x",
              AX (r), BX (r));
      }

      SET_AX (r, 0);
      break;

    case 0x09:
      lwarn ("U VIDEO - WRITE CHARACTER AND ATTRIBUTE AT CURSOR POSITION %#x\n",
             AX (r));
      break;

    case 0x0b:
      switch (BH (r)) {
        case 0:
          lwarn ("U VIDEO - SET BACKGROUND/BORDER COLOR %#x %#x\n",
                 AX (r), BX (r));
          break;

        case 1:
          lwarn ("U VIDEO - SET PALETTE %#x %#x\n", AX (r), BX (r));
          break;

        default:
          goto error;
      }
      break;

    case 0x0e:
    {
      char s[1];
      s[0] = AL (r);
      VGA_print_string_TTY (s, 1);
      break;
    }

    case 0x0f:
      SET_AH (r, MEM_VID_COLUMNS);
      SET_AL (r, MEM_VID_MODE);
      SET_BH (r, 0);
      break;

    case 0x10:
      switch (AL (r)) {
        case 0x00:              /* XXX */
        case 0x01:              /* set border */
        case 0x03:              /* toggle intensity/blinking bit */
          lwarn ("no handler for %#x\n", AX (r));
          break;

        case 0x10:
        {
          ub pal[3];
          pal[0] = DH (r);
          pal[1] = CH (r);
          pal[2] = CL (r);

          VID_set_palette (pal, BX (r), 1);
          break;
        }

        case 0x12:
        {
          int ncolors;
          int first;
          ub *pal;

          first = BX (r);
          ncolors = CX (r);
          pal = (ub *) MEM_REAL (SEGLIN (ES (r), DX (r)));

          VID_set_palette (pal, first, ncolors);
          break;
        }

        default:
          UNHANDLED_INT (r);
      }
      break;

    case 0x11:
      switch (AL (r)) {
        case 0x11:
          VID_set_font (VID_vga_def8x14_font, 0, 256, 8, 14);
          break;

        case 0x12:
          lwarn ("U VIDEO - TEXT-MODE CHARGEN - LOAD ROM 8x8"
                 "DBL-DOT PATTERNS (PS,EGA,VGA) %#x\n",
                 AX (r));
          break;

        case 0x30:
          lwarn ("U VIDEO - GET FONT INFORMATION (EGA, MCGA, VGA) %#x\n",
                 AX (r));
          break;

        case 0x02:
          lwarn ("VIDEO - TEXT-MODE CHARGEN -"
                 "LOAD ROM 8x8 DBL-DOT PATTERNS (PS,EGA,VGA)\nBL=%#x\n",
                 BL (r));
          break;

        default:
          UNHANDLED_INT (r);
      }
      break;

    case 0x12:                  /* alternate function select */
      switch (BL (r)) {
        case 0x30:              /* select vertical resolution */
        {
#if 0
          int res;
          switch (AL (r)) {
            case 0:
              res = 200;
              break;

            case 1:
              res = 350;
              break;

            case 2:
              res = 400;
              break;
          }
          VGA_set_vertical_resolution (res);
          SET_AL (r, 0x12);
#endif
          lwarn ("no handler for %#x\n", AX (r));
          break;
        }

        case 0x80:
          SET_AX (r, 0);
          break;

        default:
          UNHANDLED_INT (r);
      }
      break;

    case 0x1a:
      if (0 != AL (r))
        goto error;

      SET_AL (r, 0x1a);
      SET_BX (r, 0x0808);
      break;

    case 0x4f:
      vesa (r);
      break;

    case 0x5f:
    case 0x6f:
      lwarn ("no handler for %#x\n", AX (r));
      break;

    default:
      UNHANDLED_INT (r);
  }

  return;

 error:
  UNHANDLED_INT (r);
}
