#include <stdlib.h>

#include <inc/mem.h>
#include <inc/sig.h>
#include <inc/sel.h>
#include <inc/int.h>
#include <inc/xms.h>
#include <inc/log.h>
#include <inc/mup.h>
#include <inc/red.h>
#include <inc/dpmi.h>

#define LOG_PREFIX "I2F"
static int log_level = -2;

void I2F (INT_ARGS)
{
  if (-2 == log_level)
    log_level = MUP_get_log_level ("i2f");

  linfo ("ax = %#06x\n", AX (r));

  switch (AX (r)) {
    case 0x1000:
      lwarn ("U SHARE - INSTALLATION CHECK\n");
      break;

/* CD-ROM - INSTALLATION CHECK */
    case 0x1500:
      if (BX (r) != 0)
        goto error;

      lwarn ("U CD-ROM - INSTALLATION CHECK\n");
      SET_BX (r, 1);
      SET_CX (r, 'Z' - 'A');
      break;

/* CD-ROM - GET DRIVE DEVICE LIST */
    case 0x1501:
    {
      ub *data;

      data = (ub *) SEGLIN (ES (r), BX (r));
      lwarn ("U CD-ROM - GET DRIVE DEVICE LIST %04x:%04x\n",
             ES (r), BX (r));
      data[0] = 0xff;
      data[1] = 0xff;
      data[2] = 0xff;
      data[3] = 0xff;
      data[4] = 0xff;
      break;
    }

/* CD-ROM v2.00+ - DRIVE CHECK */
    case 0x150b:
      lwarn ("U CD-ROM v2.00+ - DRIVE CHECK %#x %#x %c\n",
             AX (r), CX (r), CX (r) + 'A');
      SET_BX (r, 0xadad);
      /* if ('Z' - 'A' == CX (r)) */
        SET_AX (r, 1);
      /* else */
        /* SET_AX (r, 0); */
      break;

/* CD-ROM v2.00+ - GET MSCDEX.EXE VERSION (GET VERSION) */
    case 0x150c:
      lwarn ("U CD-ROM v2.00+ - GET MSCDEX.EXE VERSION (GET VERSION) %#x\n",
             AX (r));
      SET_BX (r, 0x0210);
      break;

/* CD-ROM v2.00+ - GET CD-ROM DRIVE LETTERS */
    case 0x150d:
    {
      char *letters;

      lwarn ("U CD-ROM v2.00+ - GET CD-ROM DRIVE LETTERS %#x\n", AX (r));
      letters = (char *) SEGLIN (ES (r), BX (r));
      letters[0] = 'Z' - 'A';
      break;
    }

/* CD-ROM v2.00+ - GET DIRECTORY ENTRY */
    case 0x150f:
      lwarn ("U CD-ROM v2.00+ - GET DIRECTORY ENTRY %#x\n"
             "name %s\n",
             AX (r), (char *) SEGLIN (ES (r), BX (r)));
      SET_CF (r);
      SET_AX (r, 9);
      break;

/* CD-ROM v2.10+ - SEND DEVICE DRIVER REQUEST */
    case 0x1510:
      lwarn ("U CD-ROM v2.10+ - SEND DEVICE DRIVER REQUEST %#x\n",
             CX (r));
      SET_AX (r, 9);
      SET_CF (r);
      break;

/* MS Windows - WINDOWS ENHANCED MODE INSTALLATION CHECK */
    case 0x1600:
      lwarn ("U MS Windows - WINDOWS ENHANCED MODE INSTALLATION CHECK %#x\n", AX (r));
      break;

    case 0x160a:
      lwarn ("U MS Windows 3.1 - IDENTIFY WINDOWS VERSION AND TYPE %#x\n",
             AX (r));
#if 0
      SET_AX (r, 0);
      SET_BX (r, 0x0403);
      SET_CX (r, 3);
#endif
      break;

/* MS Windows, DPMI, various - RELEASE CURRENT VIRTUAL MACHINE TIME-SLICE */
    case 0x1680:
      SIG_wait ();
      break;

    case 0x1687:
    {
#ifndef QEMU
      ui dpmi_data_size;

      linfo ("DOS Protected-Mode Interface - INSTALLATION CHECK\n");

      dpmi_data_size = DPMI_RM_ASYNC_STACK_PARAS * 16;
      dpmi_data_size += DPMI_PM_ASYNC_STACK_PARAS * 16;

      SET_AX (r, 0);
      SET_BX (r, 1);
      SET_CL (r, 5);
      SET_DX (r, 90);
      SET_SI (r, dpmi_data_size);
      ES (r) = RED_SEGMENT;
      SET_DI (r, RED_SWITCH_TO_PM);
#endif
      break;
    }

    case 0x4300:
      linfo ("EXTENDED MEMORY SPECIFICATION (XMS) v2+ - INSTALLATION CHECK %#x\n",
             AX (r));
      if (XMS_active ()) {
        SET_AL (r, 0x80);
      }
      break;

    case 0x4310:
      linfo ("EXTENDED MEMORY SPECIFICATION (XMS) v2+ - GET DRIVER ADDRESS\n");
      if (XMS_active ()) {
        ES (r) = RED_SEGMENT;
        SET_BX (r, RED_XMS_JUMP);
      }
      break;

    case 0x4680:
      lwarn ("U MS Windows v3.0 - INSTALLATION CHECK %#x\n", AX (r));
      break;

    case 0x4a10:
      switch (BX (r)) {
        /* SMARTDRV v4.00+ - INSTALLATION CHECK AND HIT RATIOS */
        case 0:
          lwarn ("U SMARTDRV v4.00+ - INSTALLATION CHECK AND HIT RATIOS %#x %#x\n",
                 AX (r), BX (r));
          break;

        case 1:
          lwarn ("U SMARTDRV v4.00+ - FLUSH BUFFERS (COMMIT CACHE) %#x %#x\n",
                 AX (r), BX (r));
          break;

        default:
          UNHANDLED_INT (r);
      }
      break;

/* Novell NetWare - LOW-LEVEL API (IPX) INSTALLATION CHECK */
    case 0x7a00:
      lwarn ("U Novell NetWare - LOW-LEVEL API (IPX) INSTALLATION CHECK %#x\n",
             AX (r));
      break;

    case 0xbc00:
      if ((0x3f3f == BX (r))
          && (0 == CX (r))
          && (0 == DX (r))) {
        lwarn ("U MediaVision MVSOUND.SYS - INSTALLATION CHECK\n");
        break;
      }
      else
        UNHANDLED_INT (r);
      break;

    case 0xbc03:
      lwarn ("U MediaVision MVSOUND.SYS - GET FUNCTION TABLE POINTER\n");
      break;

    default:
      UNHANDLED_INT (r);
  }
  return;

 error:
  UNHANDLED_INT (r);
}
