#include <stdlib.h>

#include <inc/int.h>
#include <inc/mou.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "MOU"
static int log_level = -2;

void I33 (INT_ARGS)
{
  if (-2 == log_level)
    log_level = MUP_get_log_level ("i33");

  if (((3 == AX (r)) || (4 == AX (r)) || (0xb == AX (r)))
      || (log_level < LOG_DEBUG))
    linfo ("ax = %#06x\n", AX (r));
  else {
    ldebug ("ax = %#06x ", AX (r));
    INT_log_regs (r);
  }

  switch (AX (r)) {

/* MS MOUSE - RESET DRIVER AND READ STATUS */
    case 0x0000:
      SET_AX (r, 0xffff);
      SET_BX (r, 0xffff);
      break;

/* MS MOUSE v1.0+ - SHOW MOUSE CURSOR */
    case 0x0001:
      MOU_show_cursor ();
      break;

/* MS MOUSE v1.0+ - HIDE MOUSE CURSOR */
    case 0x0002:
      MOU_hide_cursor ();
      break;

/* MS MOUSE v1.0+ - RETURN POSITION AND BUTTON STATUS */
    case 0x0003: {
      int state;
      int x, y;
      static int old_state;
      static int old_x, old_y;

      state = MOU_get_state (&x, &y);

      if ((old_x == x) && (old_y == y)
          && (old_state == state)) {

        SET_BX (r, state);
        SET_CX (r, x);
        SET_DX (r, y);
        SIG_wait ();
      }
      else {
        old_state = state;
        old_x = x;
        old_y = y;

        SET_BX (r, state);
        SET_CX (r, x);
        SET_DX (r, y);
        linfo ("xpos=%d ypos=%d state=%d\n", x, y, BX (r));
      }
      break;
    }

/* MS MOUSE v1.0+ - POSITION MOUSE CURSOR */
    case 0x0004:
      MOU_set_position (CX (r), DX (r));
      break;

/* MS MOUSE v1.0+ - RETURN BUTTON PRESS DATA */
    case 0x0005:
      lwarn ("U MS MOUSE v1.0+ - RETURN BUTTON PRESS DATA %#x %#x\n",
             AX (r), BX (r));
      break;

/* MS MOUSE v1.0+ - DEFINE HORIZONTAL CURSOR RANGE */
    case 0x0007:
      MOU_set_horizontal_range (CX (r), DX (r));
      break;

/* MS MOUSE v1.0+ - DEFINE VERTICAL CURSOR RANGE */
    case 0x0008:
      MOU_set_vertical_range (CX (r), DX (r));
      break;

/* MS MOUSE v3.0+ - DEFINE TEXT CURSOR */
    case 0x000a:
      lwarn ("U MS MOUSE v3.0+ - DEFINE TEXT CURSOR\n");
      break;

/* MS MOUSE v1.0+ - READ MOTION COUNTERS */
    case 0x000b: {
      int dx, dy;

      MOU_read_motion_counters (&dx, &dy);
      SET_CX (r, dx);
      SET_DX (r, dy);
      break;
    }

/* MS MOUSE v1.0+ - DEFINE INTERRUPT SUBROUTINE PARAMETERS */
    case 0x000c:
      log ("callback %#x\n", CX (r));
      MOU_set_callback (CX (r), ES (r), EDX (r));
      break;

/* MS MOUSE v1.0+ - LIGHT PEN EMULATION ON */
    case 0x000e:
      break;

/* MS MOUSE v1.0+ - DEFINE MICKEY/PIXEL RATIO */
    case 0x000f:
      MOU_set_mickey_ratio (CX (r), DX (r));
      break;

/* MS MOUSE v5.0+ - DEFINE DOUBLE-SPEED THRESHOLD */
    case 0x0013:
      MOU_set_threshold (DX (r));
      break;

/* MS MOUSE v3.0+ - EXCHANGE INTERRUPT SUBROUTINES */
    case 0x0014: {
      int mask;
      int seg;
      ui off;

      MOU_get_callback (&mask, &seg, &off);
      MOU_set_callback (CX (r), ES (r), EDX (r));
      SET_CX (r, mask);
      ES (r) = seg;
      EDX (r) = off;
      break;
    }
/* MS MOUSE v6.0+ - RETURN DRIVER STORAGE REQUIREMENTS */
    case 0x0015:
      SET_BX (r, 0);
      break;

/* MS MOUSE v6.0+ - SET MOUSE SENSITIVITY */
    case 0x001a:
      MOU_set_mickey_ratio (BX (r), CX (r));
      MOU_set_threshold (DX (r));
      break;

/* MS MOUSE v6.0+ - RETURN MOUSE SENSITIVITY */
    case 0x001b:
      SET_BX (r, 0);
      SET_CX (r, 0);
      SET_DX (r, 0);
      break;

/* MS MOUSE v6.0+ - SET INTERRUPT RATE */
/* (Table 03176) */
/* Values for mouse interrupt rate: */
/*  00h    no interrupts allowed */
/*  01h    30 per second */
/*  02h    50 per second */
/*  03h    100 per second */
/*  04h    200 per second */
    case 0x001c:
      lwarn ("U MS MOUSE v6.0+ - SET INTERRUPT RATE %#x %#x\n",
             AX (r), BX (r));
      break;

/* MS MOUSE v6.0+ - ENABLE MOUSE DRIVER */
    case 0x0020:
      SET_AX (r, 0xffff);
      break;

/* MS MOUSE v6.0+ - SOFTWARE RESET */
    case 0x0021:
      SET_AX (r, 0xffff);
      SET_BX (r, 0xffff);
      break;

/* MS MOUSE v6.26+ - GET SOFTWARE VERSION, MOUSE TYPE, AND IRQ NUMBER */
    case 0x0024:
      SET_AX (r, 0xffff);
      break;

/* MS MOUSE v8.10+ - LCD SCREEN LARGE POINTER SUPPORT */
    case 0x0035:
      break;

/* EH? */
    case 0x53c1:
      break;

    default:
      goto error;
  }

  return;

 error:
  UNHANDLED_INT (r);
  SET_CF (r);
}

