open Types

let _ES = 0
let _CS = 1
let _DS = 3
let _SS = 2
let _FS = 4
let _GS = 5

let _AL = 0
let _CL = 1
let _DL = 2
let _BL = 3
let _AH = 4
let _CH = 5
let _DH = 6
let _BH = 7

let _AX = 0
let _CX = 1
let _DX = 2
let _BX = 3
let _SP = 4
let _BP = 5
let _SI = 6
let _DI = 7

let _EAX = 0
let _ECX = 1
let _EDX = 2
let _EBX = 3
let _ESP = 4
let _EBP = 5
let _ESI = 6
let _EDI = 7

let names8 = [|"AL";"CL";"DL";"BL";"AH"; "CH";"DH";"BH"|]
let names16 = [|"AX";"CX";"DX";"BX";"SP";"BP";"SI";"DI"|]
let names32 = [|"EAX";"ECX";"EDX";"EBX";"ESP";"EBP";"ESI";"EDI"|]
let namesXS = [|"ES";"CS";"SS";"DS";"FS";"GS"|]

exception Not_reg of string
exception Bad_size of int

let name8 index = names8.(index)
let name16 index = names16.(index)
let name32 index = names32.(index)
let nameseg index = namesXS.(index)
(*let name (size, index) =*)
(*  match size with*)
(*    8 -> name8 index*)
(*  | 16 -> name16 index*)
(*  | 32 -> name32 index*)
(*  | _ -> raise (Bad_size size)*)

let name = function
    Reg8 n -> name8 n
  | Reg16 n -> name16 n
  | Reg32 n -> name32 n
  | Seg n -> nameseg n

let size = function
    Reg8 _ -> S8
  | Reg16 _ -> S16
  | Reg32 _ -> S32
  | Seg _ -> S16

let parse s =
  let u = String.uppercase s in
  let find_reg array =
    let rec f i =
      if i = 8 then raise (Not_reg s)
      else if array.(i) = u then i else f (succ i)
    in
    f 0
  in
  let len = String.length s in
  match len with
    2 ->
      if u.[1] = 'S'
      then Seg (find_reg namesXS)
      else
	if u.[1] = 'L' or u.[1] = 'H'
	then Reg8 (find_reg names8)
	else Reg16 (find_reg names16)
  | 3 when u.[0] = 'E' -> Reg32 (find_reg names32)
  | _ -> raise (Not_reg s)
