#if ASIZE == 16
#  define A_TYPE WORD
#  define A_READ_REG READ_WORD_REG
#  define A_WRITE_REG WRITE_WORD_REG
#elif ASIZE == 32
#  define A_TYPE DWORD
#  define A_READ_REG READ_DWORD_REG
#  define A_WRITE_REG WRITE_DWORD_REG
#elif
#  error "bad ASIZE"
#endif

#if DSIZE == 8
#  define D_TYPE BYTE
#  define D_READ_MEM READ_BYTE_MEM
#  define D_READ_FLAT_MEM READ_FLAT_BYTE_MEM
#  define D_WRITE_MEM WRITE_BYTE_MEM
#  define D_WRITE_FLAT_MEM WRITE_FLAT_BYTE_MEM
#  define D_READ_REG READ_BYTE_REG
#  define D_READ_PORT READ_BYTE_PORT
#  define D_WRITE_PORT WRITE_BYTE_PORT
#  define D_WRITE_REG WRITE_BYTE_REG
#  define D_OP B_OP_NO_RESULT_BYTE
#  define INCR(f) ((-((f >> DF_SHIFT) & 1) | 1))
#elif DSIZE == 16
#  define D_TYPE WORD
#  define D_READ_MEM READ_WORD_MEM
#  define D_READ_FLAT_MEM READ_FLAT_WORD_MEM
#  define D_WRITE_MEM WRITE_WORD_MEM
#  define D_WRITE_FLAT_MEM WRITE_FLAT_WORD_MEM
#  define D_READ_REG READ_WORD_REG
#  define D_READ_PORT READ_WORD_PORT
#  define D_WRITE_PORT WRITE_WORD_PORT
#  define D_WRITE_REG WRITE_WORD_REG
#  define D_OP B_OP_NO_RESULT_WORD
#  define INCR(f) ((-((f >> DF_SHIFT) & 1) | 1) << 1)
#elif DSIZE == 32
#  define D_TYPE DWORD
#  define D_READ_MEM READ_DWORD_MEM
#  define D_READ_FLAT_MEM READ_FLAT_DWORD_MEM
#  define D_WRITE_MEM WRITE_DWORD_MEM
#  define D_WRITE_FLAT_MEM WRITE_FLAT_DWORD_MEM
#  define D_READ_REG READ_DWORD_REG
#  define D_READ_PORT READ_DWORD_PORT
#  define D_WRITE_PORT WRITE_DWORD_PORT
#  define D_WRITE_REG WRITE_DWORD_REG
#  define D_OP B_OP_NO_RESULT_DWORD
#  define INCR(f) ((-((f >> DF_SHIFT) & 1) | 1) << 2)
#elif
#  error Bad DSIZE
#endif

/* #define DEBUG */

#ifdef CMPS
do {
  A_TYPE si, di, incr;

  si = A_READ_REG (env, _REG_SI);
  di = A_READ_REG (env, _REG_DI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    rep &= 1;
    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) {
      wF = rF;
      break;
    }
    while (cx) {
      src = D_READ_FLAT_MEM (env, src_base + si);
      dst = D_READ_FLAT_MEM (env, dst_base + di);
      si += incr;
      di += incr;
      if (((src - dst) == 0) ^ rep)
        break;
      cx -= 1;
    }
    A_WRITE_REG (env, _REG_CX, cx);
  }
  else {
    src = D_READ_FLAT_MEM (env, src_base + si);
    dst = D_READ_FLAT_MEM (env, dst_base + di);
    si += incr;
    di += incr;
  }
  D_OP (cmp, src, dst, wF);
  A_WRITE_REG (env, _REG_SI, si);
  A_WRITE_REG (env, _REG_DI, di);
} while (0);
#endif

#ifdef MOVS
do {
  A_TYPE si, di, incr;

  si = A_READ_REG (env, _REG_SI);
  di = A_READ_REG (env, _REG_DI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) break;
    while (cx) {
      src = D_READ_FLAT_MEM (env, src_base + si);
      D_WRITE_FLAT_MEM (env, dst_base + di, src);
      si += incr;
      di += incr;
      cx -= 1;
#ifdef DEBUG
      env->iflags |= IF_JUMP;
      break;
#endif
    }
#ifdef DEBUG
    A_WRITE_REG (env, _REG_CX, cx);
#else
    A_WRITE_REG (env, _REG_CX, 0);
#endif
  }
  else {
    src = D_READ_FLAT_MEM (env, src_base + si);
    D_WRITE_FLAT_MEM (env, dst_base + di, src);
    si += incr;
    di += incr;
  }
  A_WRITE_REG (env, _REG_SI, si);
  A_WRITE_REG (env, _REG_DI, di);
} while (0);
#endif

#ifdef SCAS
do {
  A_TYPE di, incr;

  di = A_READ_REG (env, _REG_DI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    rep &= 1;
    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) {
      wF = rF;
      break;
    }
    while (cx) {
      dst = D_READ_FLAT_MEM (env, dst_base + di);
      di += incr;
      if (((src - dst) == 0) ^ rep)
        break;
      cx -= 1;
    }
    A_WRITE_REG (env, _REG_CX, cx);
  }
  else {
    dst = D_READ_FLAT_MEM (env, dst_base + di);
    di += incr;
  }

  D_OP (sub, src, dst, wF);
  A_WRITE_REG (env, _REG_DI, di);
} while (0);
#endif

#ifdef LODS
do {
  A_TYPE si, incr;

  si = A_READ_REG (env, _REG_SI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) {
      val = READ_BYTE_REG (env, _REG_AL);
      break;
    }
    while (cx) {
      val = D_READ_MEM (env, si);
      si += incr;
      cx -= 1;
    }
    A_WRITE_REG (env, _REG_CX, 0);
  }
  else {
    val = D_READ_MEM (env, si);
    si += incr;
  }

  A_WRITE_REG (env, _REG_SI, si);
} while (0);
#endif

#ifdef STOS
do {
  A_TYPE di, incr;

  di = A_READ_REG (env, _REG_DI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) break;
    while (cx) {
      D_WRITE_FLAT_MEM (env, dst_base + di, val);
      di += incr;
      cx -= 1;
#ifdef DEBUG
      env->iflags |= IF_JUMP;
      break;
#endif
    }
#ifdef DEBUG
    A_WRITE_REG (env, _REG_CX, cx);
#else
    A_WRITE_REG (env, _REG_CX, 0);
#endif
  }
  else {
    D_WRITE_FLAT_MEM (env, dst_base + di, val);
    di += incr;
  }

  A_WRITE_REG (env, _REG_DI, di);
} while (0);
#endif

#ifdef OUTS
do {
  A_TYPE si, incr;

  si = A_READ_REG (env, _REG_SI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) break;
    while (cx) {
      val = D_READ_FLAT_MEM (env, src_base + si);
      D_WRITE_PORT (env, rDX, val);
      si += incr;
      cx -= 1;
    };
    A_WRITE_REG (env, _REG_CX, 0);
  }
  else {
    val = D_READ_FLAT_MEM (env, src_base + si);
    D_WRITE_PORT (env, rDX, val);
    si += incr;
  }

  A_WRITE_REG (env, _REG_SI, si);
} while (0);
#endif

#ifdef INS
do {
  A_TYPE di, incr;

  di = A_READ_REG (env, _REG_DI);
  incr = INCR (rF);

  if (rep) {
    A_TYPE cx;

    cx = A_READ_REG (env, _REG_CX);
    if (0 == cx) break;
    while (cx) {
      val = D_READ_PORT (env, rDX);
      D_WRITE_FLAT_MEM (env, dst_base + di, val);
      di += incr;
      cx -= 1;
    }
    A_WRITE_REG (env, _REG_CX, 0);
  }
  else {
    val = D_READ_PORT (env, rDX);
    D_WRITE_FLAT_MEM (env, dst_base + di, val);
    di += incr;
  }

  A_WRITE_REG (env, _REG_DI, di);
} while (0);
#endif

#undef A_TYPE
#undef A_READ_REG
#undef A_WRITE_REG
#undef D_TYPE
#undef D_READ_REG
#undef D_WRITE_REG
#undef D_READ_MEM
#undef D_READ_FLAT_MEM
#undef D_WRITE_MEM
#undef D_WRITE_FLAT_MEM
#undef D_READ_PORT
#undef D_WRITE_PORT
#undef D_OP
#undef INCR

#undef DEBUG
