#ifndef __TURTLE_MOO__
#define __TURTLE_MOO__

union reg {
  uint32_t d;
  uint16_t w;
  uint8_t lh[2];
};

struct seg {
  uint32_t selector;
  uint8_t *base;
  uint32_t limit;
  uint32_t flags;
};

typedef struct X86 {
  uint8_t *pc;
  union reg regs[8];
  struct seg segs[6];

/*   struct seg tr; */
/*   struct seg ldt; */
/*   struct seg gdt; */
/*   struct seg idt; */

  uint32_t eflags;
  uint32_t eip;
  uint32_t cr[8];
  uint32_t dr[8];

  uint32_t iflags;

  int a32, o32, c32, s32, rep, seg, pending_int;
} X86;

enum {
  IF_JUMP            = 1 << 0,
  IF_INTR_DISABLED   = 1 << 1,
  IF_SEG_OVERRIDE    = 1 << 2,
  IF_INHIBIT_INTS    = 1 << 3,
  IF_EXCEPTION       = 1 << 4,
  IF_NOT_IMPLEMENTED = 1 << 5,
  IF_CPL_SHIFT       = 6,
  IF_CPL_MASK        = 3 << IF_CPL_SHIFT,
  IF_VME             = 1 << 8,
  IF_CONFIRM         = 1 << 9,
};

#ifdef HAVE_REGPARM
#define REGPARM(n) __attribute__((regparm (n)))
#else
#define REGPARM(n)
#endif

#define AP0__ (X86 *env)
#define AP1__(a1) (X86 *env, a1)
#define AP2__(a1,a2) (X86 *env, a1, a2)
#define AP3__(a1,a2,a3) (X86 *env, a1, a2, a3)

#define EP0__(n) turtle_##n(X86 *env) REGPARM(1)
#define EP1__(n,a1) turtle_##n(X86 *env, a1) REGPARM(2)
#define EP2__(n,a1,a2) turtle_##n(X86 *env, a1, a2) REGPARM(3)
#define EP__(n, ...) turtle_##n(X86 *env, __VA_ARGS__) REGPARM(3)
#define P0__(n) turtle_##n(X86 *env)
#define P__(n, ...) turtle_##n(X86 *env, __VA_ARGS__)

#define SIZE 8
#include "rwdefs.h"
#undef SIZE

#define SIZE 16
#include "rwdefs.h"
#undef SIZE

#define SIZE 32
#include "rwdefs.h"
#undef SIZE

#define SIZE 64
#include "rwdefs.h"
#undef SIZE

extern uint32_t EP1__(lea, uint8_t modrm);
extern void EP2__(write_cr, int rno, uint32_t v);
extern void EP2__(write_dr, int rno, uint32_t v);
extern void EP2__(load_segment, int seg, uint32_t sel);
extern void EP0__(not_implemented);
extern void EP0__(confirm);
extern void EP2__(exception, int intno, int err);
extern void EP1__(interrupt, int intno);
extern void turtle (X86 *env);

extern void EP1__(push16, uint16_t v);
extern void EP1__(push32, uint32_t v);
extern uint16_t EP0__(pop16);
extern uint32_t EP0__(pop32);
extern void EP0__(pushf16);
extern void EP0__(pushf32);
extern void EP0__(popf16);
extern void EP0__(popf32);
extern void EP0__(cli);
extern void EP0__(sti);

enum {_REG_AL, _REG_CL, _REG_DL, _REG_BL, _REG_AH, _REG_CH, _REG_DH, _REG_BH};
enum {_REG_AX, _REG_CX, _REG_DX, _REG_BX, _REG_SP, _REG_BP, _REG_SI, _REG_DI};
enum {_REG_EAX, _REG_ECX, _REG_EDX, _REG_EBX, _REG_ESP, _REG_EBP, _REG_ESI, _REG_EDI};
enum {_SEG_ES, _SEG_CS, _SEG_SS, _SEG_DS, _SEG_FS, _SEG_GS};

#define IP(env) (env)->eip
#define NEXT_IP(env) ((env)->pc-env->segs[_SEG_CS].base)

enum {
  CF         = 0x00000001,
  CF_SHIFT   = 0,
  PF         = 0x00000004,
  PF_SHIFT   = 2,
  AF         = 0x00000010,
  AF_SHIFT   = 4,
  ZF         = 0x00000040,
  ZF_SHIFT   = 6,
  SF         = 0x00000080,
  SF_SHIFT   = 7,
  TF         = 0x00000100,
  TF_SHIFT   = 8,
  IF         = 0x00000200,
  IF_SHIFT   = 9,
  DF         = 0x00000400,
  DF_SHIFT   = 10,
  OF         = 0x00000800,
  OF_SHIFT   = 11,
  IOPL_MASK  = 0x00003000,
  IOPL_SHIFT = 12,
  NT         = 0x00004000,
  NT_SHIFT   = 14,
  RF         = 0x00010000,
  RF_SHIFT   = 16,
  VM         = 0x00020000,
  VM_SHIFT   = 12,
  AC         = 0x00040000,
  AC_SHIFT   = 18,
  VIF        = 0x00080000,
  VIF_SHIFT  = 19,
  VIP        = 0x00100000,
  VIP_SHIFT  = 20,
  ID         = 0x00200000,
  ID_SHIFT   = 21,
  OC         = OF|CF,
  OSZP       = OF|SF|ZF|PF,
  OSZPC      = OF|SF|ZF|PF|CF,
  OSZAP      = OF|SF|ZF|AF|PF,
  OSZAPC     = OSZAP|CF,
};

#if 0
#define   CF           0x00000001U /* Carry Flag */
#define   CF_SHIFT     0U
#define   PF           0x00000004U /* Parity Flag */
#define   PF_SHIFT     2U
#define   AF           0x00000010U /* Auxillary carry Flag */
#define   AF_SHIFT     4U
#define   ZF           0x00000040U /* Zero Flag */
#define   ZF_SHIFT     6U
#define   SF           0x00000080U /* Sign Flag */
#define   SF_SHIFT     7U
#define   TF           0x00000100U /* Trap Flag */
#define   TF_SHIFT     8U
#define   IF           0x00000200U /* Interrupt Flag */
#define   IF_SHIFT     9U
#define   DF           0x00000400U /* Direction Flag */
#define   DF_SHIFT     10U
#define   OF           0x00000800U /* Overflow Flag */
#define   OF_SHIFT     11U
#define   IOPL_MASK    0x00003000U /* IOPL mask */
#define   IOPL_SHIFT   12U
#define   NT           0x00004000U /* Nested Task */
#define   NT_SHIFT     14U
#define   RF           0x00010000U /* Resume Flag */
#define   RF_SHIFT     16U
#define   VM           0x00020000U /* Virtual Mode */
#define   VM_SHIFT     12U
#define   AC           0x00040000U /* Alignment Check */
#define   AC_SHIFT     18U
#define   VIF          0x00080000U /* Virtual Interrupt Flag */
#define   VIF_SHIFT    19U
#define   VIP          0x00100000U /* Virtual Interrupt Pending */
#define   VIP_SHIFT    20U
#define   ID           0x00200000U /* CPUID detection flag */
#define   ID_SHIFT     21U

#define OC (OF|CF)
#define OSZP (OF|SF|ZF|PF)
#define OSZPC (OSZP|CF)
#define OSZAP (OF|SF|ZF|AF|PF)
#define OSZAPC (OSZAP|CF)
#endif

#define PTR uint32_t
#define BYTE uint8_t
#define SBYTE int8_t
#define WORD uint16_t
#define SWORD int16_t
#define DWORD uint32_t
#define SDWORD int32_t
#define QWORD uint64_t
#define SQWORD int64_t

#define IS_ADDRESS_SIZE_32(env) (env)->a32
#define IS_OPERAND_SIZE_32(env) (env)->o32
#define IS_STACK_32(env) (env)->s32
#define IS_CODE_32(env) (env)->c32

#define READ_BYTE_REG(env,i) (env)->regs[(i)&3].lh[(i)>>2]
#define WRITE_BYTE_REG(env,i,v) (env)->regs[(i)&3].lh[(i)>>2]=(v)

#define READ_WORD_REG(env,i) (env)->regs[i].w
#define WRITE_WORD_REG(env,i,v) (env)->regs[i].w=(v)

#define READ_DWORD_REG(env,i) (env)->regs[i].d
#define WRITE_DWORD_REG(env,i,v) (env)->regs[i].d=(v)

#define READ_BYTE_MEM turtle_read_mem8
#define READ_WORD_MEM turtle_read_mem16
#define READ_DWORD_MEM turtle_read_mem32

#define WRITE_BYTE_MEM turtle_write_mem8
#define WRITE_WORD_MEM turtle_write_mem16
#define WRITE_DWORD_MEM turtle_write_mem32

#define READ_FLAT_BYTE_MEM turtle_read_flat_mem8
#define READ_FLAT_WORD_MEM turtle_read_flat_mem16
#define READ_FLAT_DWORD_MEM turtle_read_flat_mem32

#define WRITE_FLAT_BYTE_MEM turtle_write_flat_mem8
#define WRITE_FLAT_WORD_MEM turtle_write_flat_mem16
#define WRITE_FLAT_DWORD_MEM turtle_write_flat_mem32

#define READ_SEG_BYTE_MEM turtle_read_seg_mem8
#define READ_SEG_WORD_MEM turtle_read_seg_mem16
#define READ_SEG_DWORD_MEM turtle_read_seg_mem32

#define WRITE_SEG_BYTE_MEM turtle_write_seg_mem8
#define WRITE_SEG_WORD_MEM turtle_write_seg_mem16
#define WRITE_SEG_DWORD_MEM turtle_write_seg_mem32

#define WRITE_BYTE_PORT turtle_write_port8
#define WRITE_WORD_PORT turtle_write_port16
#define WRITE_DWORD_PORT turtle_write_port32

#define READ_BYTE_PORT turtle_read_port8
#define READ_WORD_PORT turtle_read_port16
#define READ_DWORD_PORT turtle_read_port32

#define READ_SEGMENT_REG(env,i) (env)->segs[i].selector
#define WRITE_SEGMENT_REG turtle_load_segment

#define WRITE_CONTROL_REG turtle_write_cr
#define WRITE_DEBUG_REG turtle_write_dr
#define WRITE_TEST_REG turtle_write_tr

#define READ_CONTROL_REG(env,x) 0, turtle_not_implemented(env)
#define READ_DEBUG_REG(env,x) 0, turtle_not_implemented(env)
#define READ_TEST_REG(env,x) 0, turtle_not_implemented(env)

#define FETCH_BYTE turtle_fetch8
#define FETCH_WORD turtle_fetch16
#define FETCH_DWORD turtle_fetch32

#define CLI turtle_cli
#define STI turtle_sti
#define PUSH_WORD turtle_push16
#define PUSH_DWORD turtle_push32
#define POP_WORD turtle_pop16
#define POP_DWORD turtle_pop32
#define PUSH_WORD_FLAGS turtle_pushf16
#define PUSH_DWORD_FLAGS turtle_pushf32
#define POP_WORD_FLAGS turtle_popf16
#define POP_DWORD_FLAGS turtle_popf32
#define EXCEPTION turtle_exception
#define INTERRUPT turtle_interrupt

#define JUMP_NEAR_ABSOLUTE_DWORD turtle_jump_near_absolute32
#define JUMP_NEAR_RELATIVE_DWORD turtle_jump_near_relative32
#define JUMP_NEAR_ABSOLUTE_WORD turtle_jump_near_absolute16
#define JUMP_NEAR_RELATIVE_WORD turtle_jump_near_relative16

#define READ_DWORD_FLAGS(env) (env)->eflags
#define READ_WORD_FLAGS(env) ((env)->eflags&0xffff)
#define WRITE_DWORD_FLAGS(env,f) (env)->eflags=(f)
#define WRITE_WORD_FLAGS(env,f) \
(env)->eflags=((env)->eflags&~0xffff)|((f)&0xffff)

#define VME(env) ((env)->iflags&IF_VME)
#define PE(env) ((env)->cr[0]&1)
#define CPL(env) (((env)->iflags&IF_CPL_MASK)>>IF_CPL_SHIFT)

#define IS_SEG_BIG(env,seg) (((env)->segs[seg].flags>>22)&1)

static inline uint8_t P0__(fetch8)
{
  return *env->pc++;
}

static inline uint16_t P0__(fetch16)
{
  uint16_t v;

  v = *(uint16_t *) env->pc;
  env->pc += 2;
  return v;
}

static inline uint32_t P0__(fetch32)
{
  uint32_t v;

  v = *(uint32_t *) env->pc;
  env->pc += 4;
  return v;
}

static inline void P__(jump_near_absolute16, uint16_t abs)
{
  env->iflags |= IF_JUMP;
  env->eip = abs & 0xffff;
}

static inline void P__(jump_near_absolute32, uint32_t abs)
{
  env->iflags |= IF_JUMP;
  env->eip = abs;
}

static inline void P__(jump_near_relative16, int16_t rel)
{
  env->iflags |= IF_JUMP;
  env->eip = (NEXT_IP (env) + rel) & 0xffff;
}

static inline void P__(jump_near_relative32, int32_t rel)
{
  env->iflags |= IF_JUMP;
  env->eip = (NEXT_IP (env) + rel);
}

#define SET_CF(f,v) (f)|=(v)
#define SET_SF(f,v) (f)|=((v)<<SF_SHIFT)
#define SET_OF(f,v) (f)|=((v)<<OF_SHIFT)
#define SET_PF(f,v) (f)|=parity_tab[(v)&0xff]

#define MOVE_FLAGS_OC(d,s) (d)=((d)&OC)|((s)&~OC)
#define MOVE_FLAGS_OSZPC(d,s) (d)=((d)&OSZPC)|((s)&~OSZPC)
#define MOVE_FLAGS_OSZAP(d,s) (d)=((d)&OSZAP)|((s)&~OSZAP)
#define MOVE_FLAGS_OSZAPC(d,s) (d)=((d)&OSZAPC)|((s)&~OSZAPC)

#define SANITIZE_EFLAGS(f) ((f)=((f)&~0x28)|2)

extern const uint8_t parity_tab[256];

#define SET_SZP(f,s,v)                          \
do {                                            \
  f |= (0 == (v)) << ZF_SHIFT;                  \
  f |= (0 != ((v) & s)) << SF_SHIFT;            \
  f |= parity_tab[(v)&0xff];                    \
} while (0)

#include "asm.h"

#define INHIBIT_INTERRUPTS(env) (env)->iflags|=IF_INHIBIT_INTS
#define NOT_IMPLEMENTED turtle_not_implemented
#define CONFIRM turtle_confirm

#define PVI(env) 0
#define HLT NOT_IMPLEMENTED
#define ICEBP NOT_IMPLEMENTED

#define LEA turtle_lea
#define IS_MODRM_MEM(m) (0xc0!=((m)&0xc0))
#define MODRM_REG(m) (((m)>>3)&7)

#define GET_MODRM_REG(m) (((m)>>3)&7)
#define GET_MODRM_RM(m) ((m)&7)

#define INIT_MODRM(modrm) do {                  \
  if (0xc0 == ((modrm) & 0xc0)) {               \
      rm = (modrm) & 7;                         \
  }                                             \
  else {                                        \
      rm = lea (env, (modrm));                  \
  }                                             \
} while (0)

#define READ_REP_PREFIX(env) (env)->rep
#define READ_SEGMENT_BASE(env,seg) ((uint32_t)(env)->segs[seg].base)

#define HANDLE_PREFIX_ES(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_ES;                           \
} while (0)
#define HANDLE_PREFIX_CS(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_CS;                           \
} while (0)
#define HANDLE_PREFIX_SS(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_SS;                           \
} while (0)
#define HANDLE_PREFIX_DS(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_DS;                           \
} while (0)
#define HANDLE_PREFIX_FS(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_FS;                           \
} while (0)
#define HANDLE_PREFIX_GS(env) do {              \
  (env)->iflags|=IF_SEG_OVERRIDE;               \
  (env)->seg=_SEG_GS;                           \
} while (0)

#define HANDLE_PREFIX_O32(env) (env)->o32 ^= 1;
#define HANDLE_PREFIX_A32(env) (env)->a32 ^= 1;
#define HANDLE_PREFIX_LOCK(env)
#define HANDLE_PREFIX_REPNE(env) (env)->rep=0xf2;
#define HANDLE_PREFIX_REP(env) (env)->rep=0xf3;

#define DEFSEG(env) (env)->seg

#endif /* __TURTLE_MOO__ */
