type size = S8 | S16 | S32
type osize = Oany | O16 | O32
type reg = Reg8 of int | Reg16 of int | Reg32 of int | Seg of int
type zz =
  | CL
  | Rd
  | DX
  | Regb
  | M
  | Gb
  | Td
  | Ov
  | Ib
  | Jb
  | Xb
  | Yb
  | Ma
  | Ev
  | Fv
  | Rw
  | Ew
  | Mp
  | Sw
  | Regv
  | CS
  | Gv
  | Ob
  | Ap
  | DS
  | EAX
  | ES
  | Iv
  | Ms
  | Const1
  | FS
  | Jv
  | Iw
  | SS
  | GS
  | Cd
  | Xv
  | AL
  | Eb
  | Ep
  | Dd
  | Yv

type sign = Isigned | Iunsigned

type cast = Cast of sign * size | Raw

let string_of_size = function
    S8 -> "size:8"
  | S16 -> "size:16"
  | S32 -> "size:32"

let char_of_size = function
    S8 -> 'b'
  | S16 -> 'w'
  | S32 -> 'd'

let string_of_sign = function
    Isigned -> "S"
  | Iunsigned -> "U"

let char_of_sign = function
    Isigned -> 's'
  | Iunsigned -> 'u'

type v = Vw | Vd
type dep = Dep of v | Indep

type rm =
    Rm_r
  | Rm_w
  | Rm_rw
  | Rm_wf
  | Rm_reg
  | Rm_nil

type assume =
    Amem
  | Areg
  | Anone
  | Afree

type var =
    Operand of (int * zz * size * cast)
  | Reg of reg * cast
  | Flag
  | EIP

let ctypes = [|"BYTE"; "WORD"; "DWORD"; "SBYTE"; "SWORD"; "SDWORD"|]
let ctype_name code = ctypes.(code)
let ctype_code size cast =
  let sign, size =
    match cast with
      Cast(sign, size) -> sign, size
    | Raw -> Iunsigned, size
  in
  let c =
    match size with
      S8 -> 0
    | S16 -> 1
    | S32 -> 2
  in
  if sign = Isigned then c + 3 else c

