#ifndef __DOS_MOO__
#define __DOS_MOO__

#include <inc/adb.h>

enum DOS_FILE_ATTR {
  READ_ONLY    = 1,
  HIDDEN       = 2,
  SYSTEM       = 4,
  VOLUME_LABEL = 8,
  DIRECTORY    = 16,
  ARCHIVE      = 32,
  SHAREABLE    = 64
};

#define PSP_ENVIRONMENT_SEGMENT(psp)       (*(us*)(psp+0x2c))

typedef enum DOS_ERROR_CODE {
  NO_ERROR                                           = 0,
  FUNCTION_NUMBER_INVALID                            = 1,
  FILE_NOT_FOUND                                     = 2,
  PATH_NOT_FOUND                                     = 3,
  TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE           = 4,
  ACCESS_DENIED                                      = 5,
  INVALID_HANDLE                                     = 6,
  MEMORY_CONTROL_BLOCK_DESTROYED                     = 7,
  INSUFFICIENT_MEMORY                                = 8,
  MEMORY_BLOCK_ADDRESS_INVALID                       = 9,
  ENVIRONMENT_INVALID_USUALLY_32K_IN_LENGTH          = 10,
  FORMAT_INVALID                                     = 11,
  ACCESS_CODE_INVALID                                = 12,
  DATA_INVALID                                       = 13,
  RESERVED                                           = 14,
  FIXUP_OVERFLOW                                     = 14,
  INVALID_DRIVE                                      = 15,
  ATTEMPTED_TO_REMOVE_CURRENT_DIRECTORY              = 16,
  NOT_SAME_DEVICE                                    = 17,
  NO_MORE_FILES                                      = 18,
  DISK_WRITE_PROTECTED                               = 19,
  UNKNOWN_UNIT                                       = 20,
  DRIVE_NOT_READY                                    = 21,
  UNKNOWN_COMMAND                                    = 22,
  DATA_ERROR_CRC                                     = 23,
  BAD_REQUEST_STRUCTURE_LENGTH                       = 24,
  SEEK_ERROR                                         = 25,
  UNKNOWN_MEDIA_TYPE_NON_DOS_DISK                    = 26,
  SECTOR_NOT_FOUND                                   = 27,
  PRINTER_OUT_OF_PAPER                               = 28,
  WRITE_FAULT                                        = 29,
  READ_FAULT                                         = 30,
  GENERAL_FAILURE                                    = 31,
  SHARING_VIOLATION                                  = 32,
  LOCK_VIOLATION                                     = 33,
  DISK_CHANGE_INVALID_ESDI_MEDIA_ID_STRUCTURE        = 34,
  FCB_UNAVAILABLE                                    = 35,
  BAD_FAT                                            = 35,
  SHARING_BUFFER_OVERFLOW                            = 36,
  CODE_PAGE_MISMATCH                                 = 37,
  CANNOT_COMPLETE_FILE_OPERATION_EOF_OR_OUT_OF_INPUT = 38,
  INSUFFICIENT_DISK_SPACE                            = 39,
} doserr_e;

extern void     DOS_pre_init          (void);
extern void     DOS_init              (void);
extern us       DOS_largest_block     (void);
extern int      DOS_alloc             (us size, ui *linear);
extern int      DOS_resize            (ui linear, us size);
extern void     DOS_free              (ui linear);
extern ui       DOS_get_PSP           (void);
extern void     DOS_set_PSP           (us psp_seg);
extern void     DOS_create_PSP        (us psp_seg);
extern void     DOS_create_child_PSP  (ui lin, us next);
extern void     DOS_con_out           (char c);
extern int      DOS_open              (const char *path, int dmode);
extern int      DOS_create            (const char *path, int attr);
extern int      DOS_read              (int hindex, void *buf, int size);
extern int      DOS_write             (int hindex, void *buf, int size);
extern int      DOS_close             (int hindex);
extern int      DOS_seek              (int hindex, int off, int whence);
extern int      DOS_con_in            (void);
extern void     DOS_put_name          (char *name);
extern void     DOS_put_cmdline       (char *cmdline);
extern int      DOS_ioctl_get_info    (int hindex);
extern int      DOS_make_dir          (char *dos_path);
extern int      DOS_set_cur_dir       (char *dir);
extern int      DOS_get_cur_dir       (int drive, char *buf);
extern int      DOS_remove_dir        (char *dir);
extern int      DOS_get_fs_map_info   (int drive, char **map_path, char **wd);
extern int      DOS_find_first        (const char *mask, int attr);
extern int      DOS_find_next         (void);
extern void     DOS_set_DTA           (void *new_dta);
extern void   * DOS_get_DTA           (void);
extern void     DOS_put_env           (char *var);
extern void     DOS_map_default_drive (char *path);
extern int      DOS_set_drive         (int drive);
extern int      DOS_get_drive         (void);
extern int      DOS_get_highest_drive (void);
extern int      DOS_get_ndrives       (void);
extern int      DOS_get_file_attr     (const char *path);
extern int      DOS_unlink            (const char *path);
extern int      DOS_rename            (const char *oldpath, const char *newpath);
extern int      DOS_dup_handle        (int hindex);
extern void     DOS_get_list_of_lists (us *seg, us *off);
extern doserr_e DOS_get_error         (void);

#endif /* __DOS_MOO__ */
