#ifndef __REG_MOO__
#define __REG_MOO__

#include <sys/vm86.h>

#define SIGNALS_PEDAL_TO_METAL
#ifdef SIGNALS_PEDAL_TO_METAL

#  include <signal.h>
#  include <stddef.h>
#  include <assert.h>

#  define SIGEAX(r) (r)->eax
#  define SIGEBX(r) (r)->ebx
#  define SIGECX(r) (r)->ecx
#  define SIGEDX(r) (r)->edx
#  define SIGESI(r) (r)->esi
#  define SIGEDI(r) (r)->edi
#  define SIGEBP(r) (r)->ebp
#  define SIGESP(r) (r)->esp
#  define SIGEIP(r) (r)->eip
#  define SIGEFL(r) (r)->eflags
#  define SIGCS(r) (r)->cs
#  define SIGDS(r) (r)->ds
#  define SIGES(r) (r)->es
#  define SIGFS(r) (r)->fs
#  define SIGGS(r) (r)->gs
#  define SIGSS(r) (r)->ss

typedef struct sigcontext_struct regs86;
typedef unsigned long int reg86;

#  define REGS86_SZ sizeof (struct sigcontext_struct)
#  define CONTEXT_SZ sizeof (struct sigcontext_struct) + sizeof (struct _fpstate)

static inline reg86 *SIG_gregp (regs86 *r, int nreg)
{
  ui flat;

  assert (nreg < 8);
  flat = (ui) r;
  return (reg86 *) (flat + offsetof (regs86, eax) - (nreg * 4));
}

static inline us *SIG_sregp (regs86 *r, int nseg)
{
  switch (nseg) {
    case 0: return &r->es; break;
    case 1: return &r->cs; break;
    case 2: return &r->ss; break;
    case 3: return &r->ds; break;
    case 4: return &r->fs; break;
    case 5: return &r->gs; break;
    default:
      assert (0);
  }
}

#else

#  include <ucontext.h>
#  define SIGEAX(r) (*(r))[REG_EAX]
#  define SIGEBX(r) (*(r))[REG_EBX]
#  define SIGECX(r) (*(r))[REG_ECX]
#  define SIGEDX(r) (*(r))[REG_EDX]
#  define SIGESI(r) (*(r))[REG_ESI]
#  define SIGEDI(r) (*(r))[REG_EDI]
#  define SIGEBP(r) (*(r))[REG_EBP]
#  define SIGESP(r) (*(r))[REG_ESP]
#  define SIGEIP(r) (*(r))[REG_EIP]
#  define SIGEFL(r) (*(r))[REG_EFL]
#  define SIGCS(r) (*(r))[REG_CS]
#  define SIGDS(r) (*(r))[REG_DS]
#  define SIGES(r) (*(r))[REG_ES]
#  define SIGFS(r) (*(r))[REG_FS]
#  define SIGGS(r) (*(r))[REG_GS]
#  define SIGSS(r) (*(r))[REG_SS]

typedef gregset_t regs86;
typedef greg_t reg86;

#  define REGS86_SZ sizeof (gregset_t)
#  define CONTEXT_SZ sizeof (gregset_t)
#endif

#endif /* __REG_MOO__ */
