#include <inc/rtc.h>
#include <inc/int.h>
#include <inc/pic.h>
#include <inc/tid.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "RTC"
static int log_level;

static ul base;
static ul counter;
static int enabled;

void RTC_set_rate (ui hz)
{
  ldebug ("(hz=%u)\n", hz);

  if (0 == hz) {
    enabled = 0;
    counter = 0;
  }
  else {
    counter = TID_NSEC / hz;
    base = TID_get_abs_nsec ();
  }
}

void RTC_enable ()
{
  enabled = 1;
  base = TID_get_abs_nsec ();
}

void RTC_disable ()
{
  enabled = 0;
}

void RTC_run ()
{
  ul delta;

  if (enabled && counter) {
    delta = TID_get_nsec_diff (base);

    if (delta > counter) {
      base += counter;
      PIC_request_service (8);
    }
  }
}

static void IRQ8 (INT_ARGS)
{
  DEREF (r);

  PIC_eoi (1);
}

static void pic_handler (int which)
{
  if (which)
    RTC_run ();
}

void RTC_init ()
{
  log_level = MUP_get_log_level ("rtc");

  DEREF (log_level);
  INT_set (0x70, IRQ8);
  PIC_register (8, pic_handler);
}
