#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <SDL/SDL.h>

#include <inc/dos.h>
#include <inc/vbe.h>
#include <inc/asm.h>
#include <inc/mem.h>
#include <inc/vga.h>
#include <inc/vid.h>
#include <inc/mup.h>
#include <inc/log.h>

#define LOG_PREFIX "VBE"
static int log_level;

#define MI_ATTR(m) (*(us*)(m))
#define MI_WAAT(m) (*(ub*)(m+2))
#define MI_WBAT(m) (*(ub*)(m+3))
#define MI_GRAN(m) (*(us*)(m+4))
#define MI_WINS(m) (*(us*)(m+6))
#define MI_ASEG(m) (*(us*)(m+8))
#define MI_BSEG(m) (*(us*)(m+10))
#define MI_FPTR(m) (*(ui*)(mi+12))
#define MI_BPSL(m) (*(us*)(m+16))
#define MI_XRES(m) (*(us*)(m+18))
#define MI_YRES(m) (*(us*)(m+20))
#define MI_XCHR(m) (*(ub*)(m+22))
#define MI_YCHR(m) (*(ub*)(m+23))
#define MI_NPLN(m) (*(ub*)(m+24))
#define MI_BPPI(m) (*(ub*)(m+25))
#define MI_MMOD(m) (*(ub*)(m+27))
#define MI_BSIZ(m) (*(ub*)(m+28))
#define MI_NPAG(m) (*(ub*)(m+29))

#define MI_RMSK(m) (*(ub*)(m+31))
#define MI_RPOS(m) (*(ub*)(m+32))
#define MI_GMSK(m) (*(ub*)(m+33))
#define MI_GPOS(m) (*(ub*)(m+34))
#define MI_BMSK(m) (*(ub*)(m+35))
#define MI_BPOS(m) (*(ub*)(m+36))
#define MI_AMSK(m) (*(ub*)(m+37))
#define MI_APOS(m) (*(ub*)(m+38))

#define MI_DCMI(m) (*(ub*)(m+39))
#define MI_BPTR(m) (*(ui*)(m+40))
#define MI_OPTR(m) (*(ui*)(m+44))
#define MI_OSIZ(m) (*(us*)(m+48))

#define MI_SIZE 256

#define VI_SIGN(v) (*(ui*)v)
#define VI_VMIN(v) (*(ub*)(v+4))
#define VI_VMAJ(v) (*(ub*)(v+5))
#define VI_VEND(v) (*(ui*)(v+6))
#define VI_CAPS(v) (*(ui*)(v+10))
#define VI_MDLS(v) (*(ui*)(v+14))
#define VI_64KS(v) (*(us*)(v+18))

#define VI_RESV(v) ((ub*)(v+20))

#define VI2_OEMREV(v) (*(us*)(v+20))
#define VI2_VENDOR(v) (*(ui*)(v+22))
#define VI2_PRNAME(v) (*(ui*)(v+26))
#define VI2_PRREVI(v) (*(ui*)(v+30))
#define VI2_AFVERS(v) (*(us*)(v+34))
#define VI2_AFMDLS(v) (*(ui*)(v+36))
#define VI2_RESRVD(v) ((ub*)(v+40))
#define VI2_OEMBLK(v) ((ub*)(v+256))

#define FP(seg, off) ((ui)((off)+(((seg)>>4)<<16)))
#define VI_OEMSTRING FP(0xf0000,0xe094)
#define VI_MODELIST FP(0xf0000,0xe094+32)

#define VBE2 0x32454256
#define VESA 0x41534556

struct vbe_mode {
  int mattr;
  int aattr;
  int battr;
  int wingran;
  int winsize;
  int winaseg;
  int winbseg;
  int winfunc;
  int pitch;
  int xres;
  int yres;
  int xchar;
  int ychar;
  int planes;
  int bpp;
  int memmodel;
  int banksize;
  int imageplanes;
  int rmasksize;
  int rfieldpos;
  int gmasksize;
  int gfieldpos;
  int bmasksize;
  int bfieldpos;
  int amasksize;
  int afieldpos;
  int dcattr;
};

enum mode_type {
  LFB   = 1,
  PAGED = 2,
  TEXT  = 4
};

#define NO_BOARD_MODES
#ifdef NO_BOARD_MODES
static struct vbe_mode *mup_mode;
#else
static int board_modes_avail[0x200];
static ub *board_modes;
static ub cur_mode[MI_SIZE];
#endif

static char *vbe_board_modes;
static int c_vbe_mode_index;
static struct vbe_mode c_vbe_mode;
static struct vbe_mode vbe_modes[0x200];
static int vbe_mode_indices[0x200];

/* static int total_modes; */
static int use_vbe12;
static int sfks = 512;
static int vbe2;
static ub cur_win;
static us win_pages[2];
static ub *vid_buf;
static int buf_size;
static int can_mmap;
static enum mode_type vid_type;

static void *pm_interface_addr;
static ui pm_interface_size;

static struct vbe_mode *get_vbe_mode (int index)
{
  if (vbe_mode_indices[index] > 0) {
    return vbe_modes + index;
  }
  else {
    return 0;
  }
}

static void vbe_mode_set_diff (struct vbe_mode *dst)
{
  size_t i;
  ui *a;

  a = (ui *) &c_vbe_mode;

  for (i = 0; i < sizeof (c_vbe_mode) / 4; i++) {
    if (0xdeadbeef != a[i])
      ((ui *) dst)[i] = a[i];
  }
}

static void vbe_mode_post_proc ()
{
  if (c_vbe_mode_index >= 0) {
    memcpy (vbe_modes + c_vbe_mode_index, &c_vbe_mode,
            sizeof (struct vbe_mode));
  }
  else {
    int i;

    switch (c_vbe_mode_index) {
      case -1:
        for (i = 0; i < LENOFA (vbe_modes); i++) {
          vbe_mode_set_diff (vbe_modes + i);
        }
        break;

      default:
        Aldri ();
    }
  }
}

static int xstrtol (char *s, int *res)
{
  int ival;
  char *endptr;

  errno = 0;
  ival = strtol (s, &endptr, 0);
  if ((0 != endptr[0]) || (0 != errno)) {
    int err;

    err = errno;
    log ("`%s' is not an integer\n"
         "syserr %s\n",
         s, strerror (err));
    return -1;
  }

  *res = ival;
  return 0;
}

static int vbe_mode_setter (char *vmd)
{
  int index;

  vbe_mode_post_proc ();

  if (0 == strcmp (vmd, "all")) {
    index = -1;
  }
  else {
    int res;
    res = xstrtol (vmd, &index);

    if (-1 == res) {
      return -1;
    }

    if (index >= LENOFA (vbe_mode_indices)) {
      log ("mode index %d is too large, max %d\n",
           index, LENOFA (vbe_mode_indices));
    }

    if (0 == vbe_mode_indices[index])
      vbe_mode_indices[index] = 1;
  }

  if (index >= 0) {
    memcpy (&c_vbe_mode, vbe_modes + index, sizeof (struct vbe_mode));
  }
  else {
    size_t i;
    ui *dst;

    dst = (ui *) &c_vbe_mode;

    for (i = 0; i < sizeof (c_vbe_mode) / 4; i++) {
      dst[i] = 0xdeadbeef;
    }
  }

  c_vbe_mode_index = index;
  return 0;
}

static void log_vesa_mode (int n, char *mode)
{
  int size;
  int xres, yres, pitch;

  xres = MI_XRES (mode);
  yres = MI_YRES (mode);
  pitch = MI_BPSL (mode);

  if (0 == (MI_ATTR (mode) & 0x10)) {
    xres *= MI_XCHR (mode);
    yres *= MI_YCHR (mode);
    pitch = xres;
  }

  size = pitch * yres;

  DEREF (n);
  linfo ("mode %#x\n"
         "attr  %#x\n"
         "aseg  %#x\n"
         "bseg  %#x\n"
         "gran  %d\n"
         "size  %d\n"
         "res   %d, %d\n"
         "chr   %d, %d\n"
         "bppi  %d\n"
         "bpsl  %d\n"
         "red   %d, %d\n"
         "green %d, %d\n"
         "blue  %d, %d\n"
         "buf   %#x, %#x size %d\n",
         n,
         MI_ATTR (mode),
         MI_ASEG (mode),
         MI_BSEG (mode),
         MI_GRAN (mode) * 1024,
         MI_WINS (mode) * 1024,
         MI_XRES (mode), MI_YRES (mode),
         MI_XCHR (mode), MI_YCHR (mode),
         MI_BPPI (mode),
         MI_BPSL (mode),
         MI_RMSK (mode), MI_RPOS (mode),
         MI_GMSK (mode), MI_GPOS (mode),
         MI_BMSK (mode), MI_BPOS (mode),
         MI_BPTR (mode), MI_BPTR (mode) + size, size
    );
}

static struct vbe_mode *setup_mup_mode (char *mi, int index)
{
  struct vbe_mode *mode;

  mode = get_vbe_mode (index);

  if (NULL == mode)
    return 0;

  memset (mi, 0, MI_SIZE);

  MI_ATTR (mi) = mode->mattr;

  MI_BPSL (mi) = mode->pitch;
  MI_ASEG (mi) = mode->winaseg;
  MI_BSEG (mi) = mode->winbseg;
  MI_WINS (mi) = mode->winsize;
  MI_GRAN (mi) = mode->wingran;
  MI_WAAT (mi) = mode->aattr;
  MI_WBAT (mi) = mode->battr;
  MI_NPLN (mi) = mode->planes;
  MI_MMOD (mi) = mode->memmodel;
  MI_XRES (mi) = mode->xres;
  MI_YRES (mi) = mode->yres;
  MI_XCHR (mi) = mode->xchar;
  MI_YCHR (mi) = mode->ychar;
  MI_BPPI (mi) = mode->bpp;
  MI_RPOS (mi) = mode->rfieldpos;
  MI_RMSK (mi) = mode->rmasksize;
  MI_GPOS (mi) = mode->gfieldpos;
  MI_GMSK (mi) = mode->gmasksize;
  MI_BPOS (mi) = mode->bfieldpos;
  MI_BMSK (mi) = mode->bmasksize;
  MI_APOS (mi) = mode->afieldpos;
  MI_AMSK (mi) = mode->amasksize;
  MI_DCMI (mi) = mode->dcattr;
  if (0 == (MI_ATTR (mi) & 0x10))
    MI_BPTR (mi) = MEM_USER (MEM_B8000);
  else
    if (index < 0x80)
      MI_BPTR (mi) = MEM_USER (MEM_A0000);
    else
      MI_BPTR (mi) = MEM_USER ((ui) vid_buf);

  if (use_vbe12) {
    MI_ATTR (mi) &= ~0x80;
    MI_BPTR (mi) = 0;
  }
  else {
    MI_OPTR (mi) = 0xffffffff;
    MI_OSIZ (mi) = 0;
    /* MI_BPTR (mi) = 0xffffffff; */
  }
  return mode;
}

#ifdef NO_BOARD_MODES
#define setup_mode setup_mup_mode
#else
static void init_board_modes (const char *path)
{
  int i;
  FILE *f;

  board_modes = malloc (0x200 * MI_SIZE);
  if (NULL == board_modes) {
    ERRFail ("failed to get %d bytes for board modes", 0x200 * MI_SIZE);
  }

  f = fopen (path, "rb");
  if (NULL == f) {
    ERRFail ("can not open board modes `%s'", path);
  }

  linfo ("\n");
  for (i = 0; i < 0x200; i++) {
    int nread;
    int nmode;

    nread = fread (&nmode, 4, 1, f);
    if (1 != nread) {
      ERRFail ("fread failed");
    }

    if (-1 != nmode) {
      ub mode_info[MI_SIZE];

      board_modes_avail[i] = nmode;

      nread = fread (&mode_info, MI_SIZE, 1, f);
      if (1 != nread) {
        ERRFail ("fread failed");
      }

      linfo ("adding board mode %#4x %4dx%4d bpp %d\n",
             nmode,
             MI_XRES (mode_info),
             MI_YRES (mode_info),
             MI_BPPI (mode_info));
      memcpy (board_modes + i * MI_SIZE, mode_info, MI_SIZE);
    } else {
      board_modes_avail[i] = -1;
    }
  }

  fclose (f);
}

static ub *get_board_mode (int index)
{
  if (-1 == board_modes_avail[index])
    return 0;

  return board_modes + MI_SIZE * index;
}

static int setup_mode (char *mi, int index)
{
  if (board_modes) {
    ub *board_mode;

    board_mode = get_board_mode (index);
    if (NULL == board_mode) {
      return -1;
    }

    memcpy (mi, board_mode, MI_SIZE);
    if (0 == (MI_ATTR (mi) & 0x10))
      MI_BPTR (mi) = MEM_B8000;
    else
      if (index < 0x80)
        MI_BPTR (mi) = MEM_A0000;
      else
        MI_BPTR (mi) = (ui) vid_buf;
    return 0;
  }
  else {
    struct vbe_mode *mode;

    mode = setup_mup_mode (mi, index);
    if (NULL == mode) {
      return -1;
    }

    return 0;
  }
}
#endif

us VBE_get_page (ub nwin)
{
  /* Assert (PAGED != vid_type); */

  return win_pages[nwin];
}

#ifdef NO_BOARD_MODES
int VBE_get_bpsl ()
{
  if (NULL == mup_mode)
    return -1;

  return mup_mode->pitch;
}

int VBE_get_psl ()
{
  if (NULL == mup_mode)
    return -1;

  return mup_mode->xres;
}

int VBE_get_nsl ()
{
  if (NULL == mup_mode)
    return -1;

  return mup_mode->yres;
}
#else
int VBE_get_bpsl ()
{
  return MI_BPSL (cur_mode);
}

int VBE_get_psl ()
{
  return MI_XRES (cur_mode);
}

int VBE_get_nsl ()
{
  return MI_YRES (cur_mode);
}
#endif

static void copy_page (int new_nwin, int new_npage)
{
  int was_locked;
  int size;
  int off;
  int npage, nwin;
  int addr;

  was_locked = VID_lock ();

  if (can_mmap) {
    npage = new_npage;
    nwin = new_nwin;
  }
  else {
    npage = win_pages[cur_win];
    nwin = cur_win;
  }

#ifdef NO_BOARD_MODES
  if (0 == nwin)
    addr = mup_mode->winaseg * 16;
  else
    addr = mup_mode->winbseg * 16;

  size = mup_mode->winsize * 1024;
  off = npage * 1024 * mup_mode->wingran;
#else
  if (0 == nwin)
    addr = MI_ASEG (cur_mode) * 16;
  else
    addr = MI_BSEG (cur_mode) * 16;

  size = MI_WINS (cur_mode) * 1024;
  off = npage * 1024 * MI_GRAN (cur_mode);
#endif

  ldebug ("%c off %d size %d %#x\n",
          cur_win + 'A', off, size, addr);

  if (can_mmap) {
    MEM_map_address_to_video (addr, off, size);
  }
  else {
    if (size + off > buf_size) {
      size = buf_size - off;
    }
    memcpy (vid_buf + off, (void *) addr, size);
  }

  VID_maybe_unlock (was_locked);
}

us VBE_set_page (ub nwin, us npage)
{
  ldebug ("(nwin=%d, npage=%d)\n",
          nwin, npage);

  if (PAGED != vid_type) {
    lwarn ("(nwin=%d, npage=%d) but mode is not paged\n",
           nwin, npage);
    return 1;
  }

  if ((nwin == cur_win) && (win_pages[nwin] == npage)) {
    return 0;
  }

  copy_page (nwin, npage);

  cur_win = nwin;
  win_pages[nwin] = npage;
  return 0;
}

void VBE_run ()
{
  switch (vid_type) {
    case TEXT:
      VID_text_mode_draw ((void *) MEM_B8000);
      break;

    case PAGED:
      if (0 == can_mmap) {
        copy_page (0, 0);
      }
      break;

    default:
      break;
  }
}

int VBE_get_pm_interface (void **addr, ui *size)
{
  *addr = pm_interface_addr;
  *size = pm_interface_size;
  pm_interface_size = ASM_vbe_pm_interface_tab[-1];
  memcpy (pm_interface_addr, ASM_vbe_pm_interface_tab, pm_interface_size);
  return 1;
}

int VBE_get_mode_info (int nmode, char *mi)
{
  int index;
  int res;

  if (-1 == nmode)
    return 0;

  index = (nmode & 0x1ff);
#ifdef NO_BOARD_MODES
  res = NULL != setup_mode (mi, index);
#else
  res = -1 != setup_mode (mi, index);
#endif

  linfo ("(nmode=%#x, mi=%p) index %#x present %d\n",
         nmode, mi, index, res);

  if (res) {
    log_vesa_mode (index, mi);
  }

  return res;
}

void VBE_get_vesa_info (char *vi)
{
  vbe2 = (VBE2 == VI_SIGN (vi));

  linfo ("VBE %d\n", vbe2 + 1);
  VI_SIGN (vi) = VESA;
  VI_CAPS (vi) = 1;
  VI_64KS (vi) = sfks;
  VI_MDLS (vi) = VI_MODELIST;

  if (use_vbe12) {
    VI_VMIN (vi) = 2;
    VI_VMAJ (vi) = 1;
    VI_VEND (vi) = VI_OEMSTRING;
  }
  else {
    VI_VMIN (vi) = 0;
    VI_VMAJ (vi) = 2;

#if 0
    strcpy ((char *) VI2_OEMBLK (vi), (char *) 0xfe094);

    VI2_OEMREV (vi) = 0x0009;
    VI2_VENDOR (vi) = VI_OEMSTRING;
    VI2_PRNAME (vi) = VI_OEMSTRING;
    VI2_PRREVI (vi) = VI_OEMSTRING;
#else
    {
#if 0
      us *s, *d;
      us seg, off;
      ui ptr;

      seg = (ui) vi >> 4;
      off = (ui) vi - (seg << 4);
      ptr = (seg << 16) | off;

      VI_VEND (vi) = ptr + 256;
      VI_MDLS (vi) = ptr + 40;

      VI2_VENDOR (vi) = ptr + 256;
      VI2_PRNAME (vi) = ptr + 256;
      VI2_PRREVI (vi) = ptr + 256;

      s = (us *) (0xfe094 + 32);
      d = (us *) VI2_RESRVD (vi);

      Assert (total_modes - 3 >= (256 - 40) / 2);
      do {
        *d++ = *s++;
      } while (*s != 0);
#endif
    }
#endif
    VI2_AFVERS (vi) = 0;
    VI2_AFMDLS (vi) = 0;
  }
}

int VBE_set_mode (int nmode)
{
  int index;
  int addr;
  int buf_size;
  char mi[MI_SIZE];
  enum mode_type mtyp;
  struct VID_mode_info info;

  index = (nmode & 0x1ff);

  linfo ("(nmode=%#x) index %#x\n", nmode, index);

#ifdef NO_BOARD_MODES
  mup_mode = setup_mode (mi, index);
  if (NULL == mup_mode) {
    Fail ("(nmode=%#x) index %#x can not get mup_mode",
          nmode, index);
  }
#else
  if (-1 == setup_mode (mi, index)) {
    Fail ("(nmode=%#x) index %#x can not get mup_mode",
          nmode, index);
  }
  memcpy (cur_mode, mi, MI_SIZE);
#endif

  info.typ = VID_TYPE_VBE;
  info.w = MI_XRES (mi);
  info.h = MI_YRES (mi);
  info.pitch = MI_BPSL (mi);
  info.bpp = MI_BPPI (mi);
  info.rmask = ((1 << MI_RMSK (mi)) - 1) << MI_RPOS (mi);
  info.gmask = ((1 << MI_GMSK (mi)) - 1) << MI_GPOS (mi);
  info.bmask = ((1 << MI_BMSK (mi)) - 1) << MI_BPOS (mi);
  info.amask = 0;

  buf_size = MI_BPSL (mi) * info.h;
  addr = MI_BPTR (mi);

  if (0 == (MI_ATTR (mi) & 0x10)) {
    /* FIXME: for vga.emu=full */
    mtyp = TEXT;
    info.typ = VID_TYPE_VBE;
    info.w *= MI_XCHR (mi);
    info.h *= MI_YCHR (mi);
    MEM_unmap_mappings_to_video ();
  }
  else {
    if ((nmode & 0x4000) || (nmode < 0x80)) {
      mtyp = LFB;

      if (vid_buf == (void *) MEM_REAL (addr))
        MEM_unmap_mappings_to_video ();
      else
        MEM_map_address_to_video (addr, 0, buf_size);

      linfo ("%#x LFB=%#x:%d\n", nmode, addr, buf_size);
    }
    else {
      ui aaddr, baddr;
      ui wsize, wgran;

#ifdef NO_BOARD_MODES
      aaddr = mup_mode->winaseg * 16;
      baddr = mup_mode->winbseg * 16;
      wsize = mup_mode->winsize * 1024;
      wgran = mup_mode->wingran * 1024;
#else
      aaddr = MI_ASEG (cur_mode) * 16;
      baddr = MI_BSEG (cur_mode) * 16;
      wsize = MI_WINS (cur_mode) * 1024;
      wgran = MI_GRAN (cur_mode) * 1024;
#endif
      can_mmap = 0 == ((aaddr | baddr | wgran) & 0xfff);

      if (can_mmap) {
        MEM_map_address_to_video (aaddr, 0, wsize);
      }
      else {
        MEM_unmap_mappings_to_video ();
      }

      cur_win = 0;
      win_pages[0] = 0;
      win_pages[1] = 0;
      mtyp = PAGED;
    }
  }

  log_vesa_mode (index, mi);

  if (-1 == VID_create (&info)) {
    Fail ("VID_create for VBE mode %#x failed", nmode);
  }

  MEM_VID_MODE = ~0;
  vid_type = mtyp;

  switch (mtyp) {
    case TEXT:
      VID_set_palette (VID_vga_def64_palette, 0, 64);
      VID_set_font (VID_vga_def8x16_font, 0, 256, 8, 16);
      break;

    case PAGED:
      VBE_set_page (0, 0);
      VID_set_palette (VID_vga_def256_palette, 0, 256);
      break;

    case LFB:
      VID_set_palette (VID_vga_def256_palette, 0, 256);
      break;
  }

  return 1;
}

void VBE_pre_init ()
{
  static struct MUP_key vbe_mode_keys[] = {
    {"attr"              , TAG_INT, {.ip = &c_vbe_mode.mattr}},
    {"win.gran"          , TAG_INT, {.ip = &c_vbe_mode.wingran}},
    {"win.size"          , TAG_INT, {.ip = &c_vbe_mode.winsize}},
    {"wina.attr"         , TAG_INT, {.ip = &c_vbe_mode.aattr}},
    {"winb.attr"         , TAG_INT, {.ip = &c_vbe_mode.battr}},
    {"wina.seg"          , TAG_INT, {.ip = &c_vbe_mode.winaseg}},
    {"winb.seg"          , TAG_INT, {.ip = &c_vbe_mode.winbseg}},
    {"bpsl"              , TAG_INT, {.ip = &c_vbe_mode.pitch}},
    {"xres"              , TAG_INT, {.ip = &c_vbe_mode.xres}},
    {"yres"              , TAG_INT, {.ip = &c_vbe_mode.yres}},
    {"xchar"             , TAG_INT, {.ip = &c_vbe_mode.xchar}},
    {"ychar"             , TAG_INT, {.ip = &c_vbe_mode.ychar}},
    {"planes"            , TAG_INT, {.ip = &c_vbe_mode.planes}},
    {"bpp"               , TAG_INT, {.ip = &c_vbe_mode.bpp}},
    {"model"             , TAG_INT, {.ip = &c_vbe_mode.memmodel}},
    {"banksize"          , TAG_INT, {.ip = &c_vbe_mode.banksize}},
    {"imageplanes"       , TAG_INT, {.ip = &c_vbe_mode.imageplanes}},
    {"red.mask.size"     , TAG_INT, {.ip = &c_vbe_mode.rmasksize}},
    {"red.field.pos"     , TAG_INT, {.ip = &c_vbe_mode.rfieldpos}},
    {"green.mask.size"   , TAG_INT, {.ip = &c_vbe_mode.gmasksize}},
    {"green.field.pos"   , TAG_INT, {.ip = &c_vbe_mode.gfieldpos}},
    {"blue.mask.size"    , TAG_INT, {.ip = &c_vbe_mode.bmasksize}},
    {"blue.field.pos"    , TAG_INT, {.ip = &c_vbe_mode.bfieldpos}},
    {"alpha.mask.size"   , TAG_INT, {.ip = &c_vbe_mode.amasksize}},
    {"alpha.field.pos"   , TAG_INT, {.ip = &c_vbe_mode.afieldpos}},
    {"direct.color.attr" , TAG_INT, {.ip = &c_vbe_mode.dcattr}},
  };

  static struct MUP_group vbe_mode_group =
    {"vbe.mode." , LENOFA (vbe_mode_keys) , vbe_mode_keys, NULL};

  static struct MUP_key vbe_keys[] = {
    {"mode"              , TAG_SETTER , {.setter = vbe_mode_setter}},
    {"board.modes"       , TAG_STRING , {.sp = &vbe_board_modes}},
    {"use12"             , TAG_INT    , {.ip = &use_vbe12}},
    {"64ks"              , TAG_INT    , {.ip = &sfks}},
  };

  static struct MUP_group vbe_group =
    {"vbe." , LENOFA (vbe_keys) , vbe_keys, NULL};

  MUP_add_group (&vbe_mode_group);
  MUP_add_group (&vbe_group);
}

void VBE_init ()
{
  ui i;
  int nmodes;
  us *mode_list;

#ifndef NO_BOARD_MODES
  char *board_modes_path;

  board_modes_path = MUP_get_string ("vbe.board.modes");
#endif

  log_level = MUP_get_log_level ("vbe");
  vbe_mode_post_proc ();

  vid_buf = MEM_video_buffer ();
  mode_list = (us *) MEM_REAL (0xfe094 + 32);

  strcpy ((void *) MEM_REAL (0xfe094), "Minde VBE");
  nmodes = 0;

#ifndef NO_BOARD_MODES
  if (board_modes_path) {
    init_board_modes (board_modes_path);
    for (i = 0; i < 0x200; i++) {
      if (-1 != board_modes_avail[i]) {
        linfo ("adding mode %#x to list of supported modes\n", i);
        *mode_list++ = i;
        nmodes++;
      }
    }
  } else
#endif
  for (i = 0; i < 0x200; i++) {
    if (vbe_mode_indices[i] > 0) {
      linfo ("adding mode %#x to list of supported modes\n", i);
      *mode_list++ = i;
      nmodes++;
    }
  }

  linfo ("total vbe modes configued %d\n", nmodes);

  *mode_list++ = 0xffff;
  *mode_list++ = 0xffff;
  *mode_list++ = 0;

  pm_interface_addr = mode_list;
}
