#include <inc/xdt.h>
#include <inc/mup.h>
#include <inc/log.h>

#define LOG_PREFIX "LDT"
static int log_level;

unsigned short XDT_get_descriptor_access (void *descriptor)
{
  unsigned char *p;

  p = (unsigned char *) descriptor;
  return p[5] | ((p[6] & ~15) << 8);
}

void XDT_set_descriptor_access (void *descriptor, unsigned short access)
{
  unsigned char *p;

  linfo ("access = %#x\n", access);
  p = (unsigned char *) descriptor;
  p[5] = access & 0xff;
  p[6] = (p[6] & 15) | ((access >> 8) & ~15);
}

ui XDT_get_descriptor_base (void *descriptor)
{
  unsigned char *p;
  ui b00_23, b24_32;

  p = descriptor;
  b00_23 = *(unsigned int *) (p + 2) & 0xffffff;
  b24_32 = p[7];
  return b00_23 | (b24_32 << 24);
}

ui XDT_get_descriptor_limit (void *descriptor)
{
  unsigned char *p;
  ui l00_15, l16_19;

  p = descriptor;
  l00_15 = *(unsigned short *) p;
  l16_19 = p[6] & 0xf;
  return l00_15 | (l16_19 << 16);
}

void XDT_set_descriptor_base (void *descriptor, unsigned long base)
{
  unsigned char *p;

  p = descriptor;
  *(unsigned int *) (p + 2) = (base & 0xffffff) | (p[5] << 24);
  p[7] = base >> 24;
}

void XDT_set_descriptor_limit (void *descriptor, unsigned int limit)
{
  unsigned char *p;

  p = descriptor;
  *(unsigned short *) p = limit & 0xffff;
  p[6] = (p[6] & ~0xf) | ((limit >> 16) & 0xf);
}

void XDT_init ()
{
  log_level = MUP_get_log_level ("xdt");
}
